import numpy as np
import os
import sys
sys.path.insert(0, '/Users/macuser/Roman/corgidrp')

import corgidrp.data

import astropy
import astropy.io.ascii as ascii
from astropy.coordinates import SkyCoord

import pyklip.fakes as fakes

import scipy.ndimage as ndi
import scipy.optimize as optimize

def centroid(frame):
    """
    Finds the center coordinates for a given frame.

    Args:
        frame (np.ndarray): 2D array to compute centering

    Returns:
        tuple:
            xcen (float): X centroid coordinate
            ycen (float): Y centroid coordinate

    """
    y, x = np.indices(frame.shape)
    
    ycen = np.sum(y * frame)/np.sum(frame)
    xcen = np.sum(x * frame)/np.sum(frame)
    
    return xcen, ycen


def centroid_with_roi(frame, roi_radius=5, centering_initial_guess=None):
    """
    Finds the centroid in a sub-region around a given initial guess (or the brightest pixel if no guess is provided).

    Args:
        frame (np.ndarray): 2D array to compute centering.
        roi_radius (int or float): Half-size of the box around the initial guess or brightest pixel.
        centering_initial_guess (tuple or None, optional): (x_init, y_init) as initial guess for centroiding.
                                                           If None, defaults to the brightest pixel.

    Returns:
        tuple:
            xcen (float): X centroid coordinate.
            ycen (float): Y centroid coordinate.
    """

    # 1) Unpack initial guess or fall back to brightest pixel
    if centering_initial_guess is not None and None not in centering_initial_guess:
        peak_x, peak_y = int(round(centering_initial_guess[0])), int(round(centering_initial_guess[1]))
    else:
        peak_y, peak_x = np.unravel_index(np.nanargmax(frame), frame.shape)

    # 2) Define the subarray (region of interest) around the peak
    y_min = max(0, peak_y - roi_radius)
    y_max = min(frame.shape[0], peak_y + roi_radius + 1)
    x_min = max(0, peak_x - roi_radius)
    x_max = min(frame.shape[1], peak_x + roi_radius + 1)
    sub_frame = frame[y_min:y_max, x_min:x_max]

    # 3) Create index arrays for sub_frame, offset so they match the full-frame coords
    y_indices, x_indices = np.indices(sub_frame.shape)
    y_indices += y_min
    x_indices += x_min

    # 4) Compute flux-weighted centroid in the subarray
    total_flux = np.sum(sub_frame)
    if total_flux == 0:
        # Edge case: empty or zero frame
        return peak_x, peak_y

    xcen = np.sum(x_indices * sub_frame) / total_flux
    ycen = np.sum(y_indices * sub_frame) / total_flux

    return xcen, ycen

def shift_psf(frame, dx, dy, flux, fitsize=10, stampsize=10):
    """
    Creates a template for matching psfs.

    Args:
        frame (np.ndarray): 2D array 
        dx (float): Value to shift psf in 'x' direction
        dy (float): Value to shift psf in 'y' direction
        flux (float): Peak flux value
        fitsize (float): (Optional) Width of square on frame to fit psf to
        stampsize (float): (Optional) Width of psf stamp to create

    Returns:
        ndi..map_coordinates: New coordinates for a shifted psf

    """
    ystamp, xstamp = np.indices([fitsize, fitsize], dtype=float)
    xstamp -= fitsize//2
    ystamp -= fitsize//2
    
    xstamp += stampsize//2 + dx
    ystamp += stampsize//2 + dy
    
    return ndi.map_coordinates(frame * flux, [ystamp, xstamp], mode='constant', cval=0.0).ravel()
    
def measure_offset(frame, xstar_guess, ystar_guess, xoffset_guess, yoffset_guess, guessflux=1, rad=5, stampsize=10):
    """
    Computes the relative offset between stars.
    
    Args:
        frame (np.ndarray): 2D array of data
        xstar_guess (float): Estimate of first star 'x' coordinate
        ystar_guess (float): Estimate of first star 'y' coordinate
        xoffset_guess (float): Estimate of 'x' direction offset between stars
        yoffset_guess (float): Estimate of 'y' direction offset between stars
        guessflux (float): (Optional) Peak flux of first star
        rad (int): (Optional) Radius around first star to compute centroid on
        stampsize (float): (Optional) Width of square psf stamp

    Returns:
        binary_offset (np.array): List of [x,y] offsets in respective directions
        fit_errs (np.array): Array of [x,y] fitting errors
    """
    #### Centroid on location of star ###
    yind = int(ystar_guess)
    xind = int(xstar_guess)
        
    ymin = yind - rad
    ymax = yind + rad + 1
    xmin = xind - rad
    xmax = xind + rad + 1
    # check bounds
    if ymin < 0:
        ymin = 0
        ymax = 2*rad + 1
    if xmin < 0:
        xmin = 0
        xmax = 2*rad + 1
    if ymax > frame.shape[0]:  # frame is the entire image here
        ymax = frame.shape[0]
        ymin = frame.shape[0] - 2 * rad - 1
    if xmax > frame.shape[1]:
        xmax = frame.shape[1] 
        xmin = frame.shape[1] - 2 * rad - 1
        
    cutout = frame[ymin:ymax, xmin:xmax]
    
    xstar, ystar = centroid(cutout)
    xstar += xmin
    ystar += ymin
    
    ### Create a PSF stamp ###
    ystamp, xstamp = np.indices([stampsize, stampsize], dtype=float)
    xstamp -= float(stampsize//2)
    ystamp -= float(stampsize//2)
    xstamp += xstar
    ystamp += ystar
    
    stamp = ndi.map_coordinates(frame, [ystamp, xstamp])
    
    ### Create a data stamp ###
    fitsize = stampsize
    ydata,xdata = np.indices([fitsize, fitsize], dtype=float)
    xdata -= fitsize//2
    ydata -= fitsize//2
    xdata += xstar + xoffset_guess
    ydata += ystar + yoffset_guess
    
    data = ndi.map_coordinates(frame, [ydata, xdata])
    
    ### Fit the PSF to the data ###
    popt, pcov = optimize.curve_fit(shift_psf, stamp, data.ravel(), p0=(0,0,guessflux), maxfev=2000)
    tinyoffsets = popt[0:2]
    fit_errs = np.sqrt([pcov[0,0], pcov[1,1], pcov[2,2]])

    binary_offset = [xoffset_guess - tinyoffsets[0], yoffset_guess - tinyoffsets[1]]

    return binary_offset, fit_errs

def compute_combinations(iteration, r=2):
    """ 
    Rough equivalivent to itertools.combinations function to create all r-length combinations from a given array.

    Args:
        iteration (np.array): Array from which to create combinations of elements
        r (int): (Optional) Length of combinations (default: 2)

    """
    inds = np.indices(np.shape(iteration))
    pool = tuple(inds[0])
    n = len(pool)
    if r > n:
        return
    
    indices = list(range(r))
    yield tuple(pool[i] for i in indices)
    while True:
        for i in reversed(range(r)):
            if indices[i] != i + n - r:
                break
        else:
            return
        indices[i] += 1
        for j in range(i+1, r):
            indices[j] = indices[j-1] + 1
        yield tuple(pool[i] for i in indices)

def angle_between(pos1, pos2):
    """ 
    Used to find the angle from North counterclockwise between two sources in an image.
    
    Args:
        pos1 (tuple): Position of the first target
        pos2 (tuple): Position of the second target
    
    Returns:
        angle (float): Angle [deg] between two sources, from north going counterclockwise
    """
    
    xdif = pos2[0] - pos1[0]
    ydif = pos2[1] - pos1[1]  

    if xdif<0:
        if ydif<0:
            angle = np.pi - np.arctan(xdif/ydif)
        else:
            angle = - np.arctan(xdif/ydif)
    else:
        if ydif<0:
            angle = - np.arctan(xdif/ydif) + np.pi
        else: 
            angle = (2*np.pi) - np.arctan(xdif/ydif)
            
    return angle * 180/np.pi


def get_polar_dist(seppa1,seppa2):
    """Computes the linear distance between two points in polar coordinates.

    Args:
        seppa1 (tuple): Separation (in any units) and position angle (in degrees) of the first point.
        seppa2 (tuple): Separation (in same units as above) and position angle (in degrees) of the second point.

    Returns:
        float: Distance between the two points in the input separation units.
    """
    sep1, pa1 = seppa1
    sep2, pa2 = seppa2

    return np.sqrt(sep1**2 + sep2**2 - (2 * sep1 * sep2 * np.cos((pa1-pa2)*np.pi/180.)))


def seppa2dxdy(sep_pix,pa_deg):
    """Converts position in separation (pixels from some reference center) and position angle 
    (counterclockwise from north) to separation in x and y pixels from the center.

    Args:
        sep_pix (float or np.array): Separation in pixels
        pa_deg (float or np.array): Position angle in degrees (counterclockwise from North)

    Returns:
        np.array: array of shape (2,) containing delta x and delta y in pixels from the center 
    """
    dx = -sep_pix * np.sin(pa_deg * np.pi/180.)
    dy = sep_pix * np.cos(pa_deg * np.pi/180.)

    return np.array([dx, dy])


def seppa2xy(sep_pix,pa_deg,cenx,ceny):
    """Converts position in separation (pixels from some reference center) and position angle 
    (counterclockwise from north) to separation in x and y pixels from the center.

    Args:
        sep_pix (float or np.array): Separation in pixels
        pa_deg (float or np.array): Position angle in degrees (counterclockwise from North)
        cenx (float): X location of center reference pixel. (0,0) is center of bottom left pixel
        ceny (float): Y location of center reference pixel. (0,0) is center of bottom left pixel

    Returns:
        np.array: x and y pixel location. (0,0) is center of bottom left pixel.
    """
    dx, dy = seppa2dxdy(sep_pix,pa_deg)

    x = dx + cenx
    y = dy + ceny

    return np.array([x, y])


def find_source_locations(image_data, threshold=10, fwhm=7, mask_rad=1):
    ''' 
    Used to find to [pixel, pixel] locations of the sources in an image
    
    Args:
        image_data (numpy.ndarray): 2D array of image data
        threshold (int): Number of stars to find (default: 100)
        fwhm (float): Full width at half maximum of the stellar psf (default: 7, ~fwhm for a normal distribution with sigma=3)
        mask_rad (int): Radius of mask for stars [in fwhm] (default: 1)
    
    Returns:
        sources (astropy.table.Table): Astropy table with columns 'x', 'y' as pixel locations
    
    '''
    # create a place to store the location arrays and use a copy of the input image
    image = np.copy(image_data)
    image_shape = np.shape(image)
    xs = np.empty(threshold) * np.nan
    ys = np.empty(threshold) * np.nan
    
    fwhm = fwhm * mask_rad
        
    i = 0
    while i < threshold:
        ind = np.where(image == np.nanmax(image))
        if len(ind) > 2:
            ind = ind[0:2]

        # record the location of the star
        image_y, image_x = np.shape(image)
        x = ind[1][0]
        y = ind[0][0]
        
        if i > 2:
            if (ys[i-1] == y and xs[i-1] == x):
                break
            
        ys[i] = y
        xs[i] = x
    
        # mask out the image at this location
        if (x - fwhm) < 0:  # left edge
            startx = 0
            endx = x + int(fwhm) + 1
            
            if (y - fwhm) < 0:  # bottom left corner
                starty = 0
                endy = y + int(fwhm) + 1
                
            elif (y + fwhm) > (image_y - 1):  # upper left corner
                starty = y - int(fwhm)
                endy = image_y

            else:
                starty = y - int(fwhm)
                endy = y + int(fwhm) + 1
        
        elif (x + fwhm) > (image_x - 1):  # right edge
            startx = x - int(fwhm)
            endx = image_x
            
            if (y - fwhm) < 0:  # bottom right corner
                starty = 0
                endy = y + int(fwhm) + 1
                
            elif (y + fwhm) > (image_y - 1):  # upper right corner
                starty = y - int(fwhm)
                endy = image_y

            else:
                starty = y - int(fwhm)
                endy = y + int(fwhm) + 1

        else:
            startx = x - int(fwhm)
            endx = x + int(fwhm) + 1
            
            if (y - fwhm) < 0:  # bottom
                starty = 0
                endy = y + int(fwhm) + 1
                
            elif (y + fwhm) > (image_y - 1):  # upper
                starty = y - int(fwhm)
                endy = image_y

            else:
                starty = y - int(fwhm)
                endy = y + int(fwhm) + 1
        
        stamp =  image[starty: endy, startx: endx]

        image[starty: endy, startx: endx] = np.zeros(np.shape(stamp))
        
        i += 1        
    # record the locations in an astropy.table
    sources = astropy.table.Table()
    sources['x'] = xs[~np.isnan(xs)]
    sources['y'] = ys[~np.isnan(ys)]

    fit_xs = np.empty(len(xs[~np.isnan(xs)]))
    fit_ys = np.empty(len(ys[~np.isnan(ys)]))

    # fit a gaussian to the guess to find true pixel pos
    # pad the image to fit stars along the edge

    pad = 25
    full_frame = np.zeros([image_shape[1]+(pad*2), image_shape[0]+(pad*2)])
    full_frame[pad:-pad, pad:-pad] = image_data
    fit_gauss_image = np.copy(full_frame)

    for i, (gx, gy) in enumerate(zip(xs[~np.isnan(xs)], ys[~np.isnan(ys)])):
        pf, fw, x, y = fakes.gaussfit2d(frame= fit_gauss_image, xguess= gx+pad, yguess=gy+pad)
        fit_xs[i] = x - pad
        fit_ys[i] = y - pad

    found_sources = astropy.table.Table()
    found_sources['x'] = fit_xs
    found_sources['y'] = fit_ys
    
    return found_sources

def match_sources(image, sources, field_path, comparison_threshold=50, rad=0.012, platescale_guess=21.8, platescale_tol=0.1):
    ''' 
    Function to find the corresponding RA/Dec positions to image sources, given a particular field.

    Args:
        image (corgidrp.data.Image): Image data as a corgidrp Image object
        sources (astropy.table.Table): Astropy table with columns 'x', 'y' as pixel locations of sources to match
        field_path (str): Full path to directory with search field data (ra, dec, vmag, etc.)
        comparison_threshold (int): How many stars in the field to consider for the initial match
        rad (float): The radius [deg] around the target coordinate for creating a subfield to match image sources to
        platescale_guess (float): An initial guess for the platescale value (default: 21.8 [mas/ pixel])
        platescale_tol (float): A tolerance for finding source matches within a fraction of the initial plate scale guess (default: 0.5)

    Returns:
        matched_sources (astropy.table.Table): Astropy table with columns 'x','y','RA', 'DEC' as pixel locations and corresponding sky positons
        
    '''
    # ensure the search field data has the proper column names
    field = ascii.read(field_path)
    if 'RA' and 'DEC' and 'VMAG' not in field.colnames:
        raise ValueError('field data must have column names [\'RA\',\'DEC\', \'VMAG\']')

    # gather the pixel locations for the (3) brightest sources in the image
    source1, source2, source3 = sources[0], sources[1], sources[2]

    # define the side length to perimeter ratio for the triangle made from these sources [pixels]
    l1, l2, l3 = np.sqrt(np.power(source1['x'] - source2['x'], 2) + np.power(source1['y'] - source2['y'], 2)), np.sqrt(np.power(source2['x'] - source3['x'], 2) + np.power(source2['y'] - source3['y'], 2)), np.sqrt(np.power(source3['x'] - source1['x'], 2) + np.power(source3['y'] - source1['y'], 2))
    perimeter = l1 + l2 + l3

    # the shortest to longest sides get reordered to l1, l2, l3
    l1, l2, l3 = np.sort([l1, l2, l3])

    a, b, c = l1/perimeter, l2/perimeter, l3/perimeter

    # define a search field and load in RA, DEC, Vmag
    field = ascii.read(field_path)
    target = image.pri_hdr['RA'], image.pri_hdr['DEC']
    
    ymid, xmid = image.data.shape   # fit gaussian to find target x,y location (assuming near center)
    pf, fw, targetx, targety = fakes.gaussfit2d(frame= image.data, xguess= (xmid-1)//2, yguess= (ymid-1)//2)

    target_skycoord = SkyCoord(ra= target[0], dec= target[1], unit='deg')
    subfield = field[((field['RA'] >= target[0] - rad) & (field['RA'] <= target[0] + rad) & (field['DEC'] >= target[1] - rad) & (field['DEC'] <= target[1] + rad))]

    bright_order_subfield = subfield.copy()
    bright_order_subfield.sort(keys='VMAG')
    brightest_field = bright_order_subfield[:comparison_threshold]

    # create all combinations of triangles with the brightest field sources
    combos = list(compute_combinations(range(len(brightest_field)), r=3))

    #a_prime, b_prime, c_prime = np.empty(len(combos)), np.empty(len(combos)), np.empty(len(combos))
    skycoords = SkyCoord(ra= brightest_field['RA'], dec= brightest_field['DEC'], unit='deg')
    field_side_lengths = np.empty((len(combos), 3))
    best_sky_ind = np.nan
    smallest_lsq = 1e10
    best_ind = np.nan

    for i, ind in enumerate(combos):
        j, k, l = ind
        s1, s2, s3 = skycoords[j], skycoords[k], skycoords[l]

        len1, len2, len3 = np.sort([s1.separation(s2).mas, s2.separation(s3).mas, s3.separation(s1).mas])

        field_side_lengths[i] = np.array([len1, len2, len3])
        perimeter = len1 + len2 + len3
        ap, bp, cp = len1/perimeter, len2/perimeter, len3/perimeter

        # make sure plate scale is within tolerance of the guess or else discard this possibility
        if ((len1 / l1) > platescale_guess* (1 + platescale_tol)) or ((len2 / l2) > platescale_guess* (1 + platescale_tol)) or ((len3 / l3) > platescale_guess* (1 + platescale_tol)):
            ap, bp, cp = 0, 0, 0
        if ((len1 / l1) < platescale_guess* (1 - platescale_tol)) or ((len2 / l2) < platescale_guess* (1 - platescale_tol)) or ((len3 / l3) < platescale_guess* (1 - platescale_tol)):
            ap, bp, cp = 0, 0, 0

        # find the best fit to the brightest image triangle
        lstsq = (a - ap)**2 + (b - bp)**2 + (c - cp)**2
        if lstsq < smallest_lsq:
            smallest_lsq = lstsq
            best_ind = i
            best_sky_ind = ind

    # now use the side length to separations with best fit triangle to define a pseudo plate scale
    best_l1, best_l2, best_l3 = field_side_lengths[best_ind]
    initial_platescale = np.mean(np.array([best_l1 / l1, best_l2 / l2, best_l3 / l3]))  # [deg/mas]

    # find pseudo north angle from difference in triangle rotations from the target value
    # using found target pixel
  
    # rot_image = np.array([angle_between(((xmid-1) //2, (ymid-1) //2), (s['x'], s['y'])) for s in [source1, source2, source3]])
    rot_image = np.array([angle_between((targetx, targety), (s['x'], s['y'])) for s in [source1, source2, source3]])
    rot_field = np.array([target_skycoord.position_angle(t).deg for t in skycoords[[best_sky_ind]]])

    initial_northangle = np.abs(np.mean(rot_field - rot_image))

    # make a new image header with the pseudo platescale and north angle to find matchings
    # allow for some error window and assign the closest star to each source
    vert_ang = np.radians(initial_northangle)
    pc = np.array([[-np.cos(vert_ang), np.sin(vert_ang)], [np.sin(vert_ang), np.cos(vert_ang)]])
    cdmatrix = pc * (initial_platescale * 0.001) / 3600.

    new_hdr = {}
    new_hdr['CD1_1'] = cdmatrix[0,0]
    new_hdr['CD1_2'] = cdmatrix[0,1]
    new_hdr['CD2_1'] = cdmatrix[1,0]
    new_hdr['CD2_2'] = cdmatrix[1,1]
    # new_hdr['CRPIX1'] = (np.shape(image.data)[1]-1) // 2
    # new_hdr['CRPIX2'] = (np.shape(image.data)[0]-1) // 2
    new_hdr['CRPIX1'] = targetx
    new_hdr['CRPIX2'] = targety
    new_hdr['CTYPE1'] = 'RA---TAN'
    new_hdr['CTYPE2'] = 'DEC--TAN'
    new_hdr['CDELT1'] = (initial_platescale * 0.001) / 3600.
    new_hdr['CDELT2'] = (initial_platescale * 0.001) / 3600.
    new_hdr['CRVAL1'] = target[0]
    new_hdr['CRVAL2'] = target[1]
    w = astropy.wcs.WCS(new_hdr)

    # transform the subfield skycoords to pixel locations
    subfield_skycoords = SkyCoord(ra= subfield['RA'], dec= subfield['DEC'], unit='deg')
    x_sky_to_pix, y_sky_to_pix = astropy.wcs.utils.skycoord_to_pixel(subfield_skycoords, wcs=w)
    # restrict to only the sources that fall within 1024 x 1024 pixels
    image_inds = ((x_sky_to_pix >= 0) & (x_sky_to_pix <= 1024) & (y_sky_to_pix >= 0) & (y_sky_to_pix <= 1024))
    x_predict = x_sky_to_pix[image_inds]
    y_predict = y_sky_to_pix[image_inds]
    subfield_skycoords_in_image = subfield_skycoords[image_inds]

    # for each source in the image, find the closest x, y predicted position and record the corresponding RA, DEC in the table
    matched_ra, matched_dec = np.empty(len(sources)), np.empty(len(sources))

    for i, (x, y)in enumerate(zip(sources['x'], sources['y'])):
        lst = 1e6
        i_match = np.nan
        for ii, (xp, yp) in enumerate(zip(x_predict, y_predict)):
            sq = (x - xp)**2 + (y - yp)**2
            if sq < lst:
                lst = sq
                i_match = ii
        matched_ra[i] = subfield_skycoords_in_image[i_match].ra.value
        matched_dec[i] = subfield_skycoords_in_image[i_match].dec.value
    
    matched_image_to_field = astropy.table.Table()
    matched_image_to_field['x'] = sources['x']
    matched_image_to_field['y'] = sources['y']
    matched_image_to_field['RA'] = matched_ra
    matched_image_to_field['DEC'] = matched_dec

    # append each x,y,RA,DEC string to the fits ext_hdr to save the source matches for reference
    for source in np.arange(len(matched_image_to_field)):
        key = 'star' + str(source + 1)

        string = str(matched_image_to_field[source]['x'])
        for col in ['y','RA','DEC']:
            string += ',' + str(matched_image_to_field[source][col])

        if key not in image.ext_hdr:
            image.ext_hdr[key] = string


    return matched_image_to_field

def fit_distortion_solution(params, fitorder, platescale, rotangle, pos1, meas_offset, sky_offset, meas_errs):
    '''
    Cost function used to fit the legendre polynomials for distortion mapping.

    Args:
        params (list): List of the x and y legendre polynomial coefficients
        fitorder (int): The degree of legendre polynomial being used
        platescale (float): The platescale of the image
        rotangle (float): The north angle of the image
        pos1 (np.array): A (2 x N) array of (x, y) pixel positions for the first star in N pairs
        meas_offset (np.array): A (2 x N) array of (x, y) pixel offset from the first star position for N star pairs
        sky_offset (np.array): A (2 x N) array of (sep, pa) true sky offsets in [mas] and [deg] from the first star position for N pairs 
        meas_errs (np.array): A (2 x N) array of (x, y) pixel errors in measured offsets from the first star position for N pairs

    Returns:
        residuals (list): List of residuals between true and measured star positions
    '''
    
    fitparams = (fitorder + 1)**2
    
    leg_params_x = np.array(params[:fitparams])  # the first half of params are for x fitting
    leg_params_x = leg_params_x.reshape(fitorder+1, fitorder+1)

    leg_params_y = np.array(params[fitparams:]) # the last half are for y fitting
    leg_params_y = leg_params_y.reshape(fitorder+1, fitorder+1)

    total_orders = np.arange(fitorder+1)[:,None] + np.arange(fitorder+1)[None,:]  # creating a 4 x 4 matrix of order numbers (?)

    leg_params_x = leg_params_x / 500**(total_orders)  # making the coefficients sufficiently large for fitting (or else ~0)
    leg_params_y = leg_params_y / 500**(total_orders)

    residuals = []

    binary_offsets = np.copy(meas_offset).T
    star1_pos = np.copy(pos1).T
    # center to center of detector
    star1_pos[:,0] -= 511.        # because leg coeffs are defined around (0,0)
    star1_pos[:,1] -= 511.        # make a new param in the function to pass in detector shape?

    # derive star2 position
    star2_pos = star1_pos + binary_offsets

    # undistort x and y for both star positions
    star1_pos_corr = np.copy(star1_pos)
    star1_pos_corr[:,0] = np.polynomial.legendre.legval2d(star1_pos[:,0], star1_pos[:,1], leg_params_x)
    star1_pos_corr[:,1] = np.polynomial.legendre.legval2d(star1_pos[:,0], star1_pos[:,1], leg_params_y)
    star2_pos_corr = np.copy(star2_pos)
    star2_pos_corr[:,0] = np.polynomial.legendre.legval2d(star2_pos[:,0], star2_pos[:,1], leg_params_x)
    star2_pos_corr[:,1] = np.polynomial.legendre.legval2d(star2_pos[:,0], star2_pos[:,1], leg_params_y)

    # translate offsets from [mas] sep, pa to [pixel] sep, pa
    sky_sep = sky_offset[0]
    # this cant be negative by definition, but is defined from north to east while corr_pa starts at (1,0) 
    sky_pa = sky_offset[1]
    
    true_offset_sep = sky_sep / platescale
    true_offset_pa = sky_pa - rotangle  # this is in degrees
    
    # translate star_pos_corr from x, y to sep, pa
    corr_offset_x = star2_pos_corr[:,0] - star1_pos_corr[:,0] 
    corr_offset_y = star2_pos_corr[:,1] - star1_pos_corr[:,1]
    
    corr_offset_sep = np.sqrt(corr_offset_x**2 + corr_offset_y**2)
    corr_offset_pa = np.degrees(np.arctan2(-corr_offset_x, corr_offset_y))
    # corr_offset_pa = np.degrees(np.arctan2(corr_offset_y, corr_offset_x))  # this is in degrees

    res_sep = corr_offset_sep - true_offset_sep  # this is in pixels
    
    res_pa_arctan_num = np.sin(np.radians(corr_offset_pa - true_offset_pa)) 
    res_pa_arctan_denom = np.cos(np.radians(corr_offset_pa - true_offset_pa))
    res_pa = np.arctan(res_pa_arctan_num / res_pa_arctan_denom) # this is in radians

    # translate pixel error in measurement to sep pa errs
    # sep_err = np.sqrt(meas_errs[0]**2 + meas_errs[1]**2) # this is in 
    sep_err = np.mean([meas_errs[0], meas_errs[1]], axis=0)
    # should be the mean of two errors also 

    ## can assume equal errors in cartesian 

    # pa_err = np.arctan2(-meas_errs[0],  meas_errs[1]) # this is in radians
    # pa_err = np.arctan2(meas_errs[1],  meas_errs[0]) # this is in radians
    pa_err = sep_err / true_offset_sep
    # should be avg of errors divided by sep --> pa err is delta arclength ~~ np.avg(m[1], m[0]) /sep
    # or fit x y res instead of sep pa

    res_sep /= sep_err
    res_pa /= pa_err
    ## translate this back to pixels x/y
    res_x = np.cos(res_pa) * res_sep # this is in pixels now
    res_y = np.sin(res_pa) * res_sep # this is in pixels now
    ## have to compute x y first and then subtract for actual res_x, res_y

    residuals = np.append(residuals, np.array([res_sep, res_pa]).ravel())

    return residuals

def compute_platescale_and_northangle(image, source_info, center_coord, center_radius=0.9):
    """
    Used to find the platescale and north angle of the image. Calculates the platescale for each pair of stars in the image
    and returns the averged platescale. Calculates the north angle for pairs of stars with the center target
    and returns the averged north angle.
    
    Args:
        image (numpy.ndarray): 2D array of image data 
        source_info (astropy.table.Table): Estimated pixel positions of sources and true sky positions, must have column names 'x', 'y', 'RA', 'DEC'
        center_coord (tuple):
            (float): RA coordinate of the target pointing
            (float): Dec coordinate of the target pointing
        center_radius (float): Percent of the image radius used to crop the image and compute plate scale and north angle from (default: 1 -- ie: the full image is used)

    Returns:
        platescale (float): Platescale [mas/pixel]
        north_angle (float): Angle between image north and true north [deg]
        
    """

    # load in the image data and source information
    if type(image) != np.ndarray:
        raise TypeError('Image must be 2D numpy.ndarray')

    if type(source_info) != astropy.table.Table:
        raise TypeError('source_info must be an astropy table with columns \'x\',\'y\',\'RA\',\'DEC\'')
    else:
        guesses = source_info
        skycoords = SkyCoord(ra = guesses['RA'], dec= guesses['DEC'], unit='deg', frame='icrs')

    # translate the center_coord param into a skycoord
    if type(center_coord) != tuple:
        raise TypeError('center_coord must be a tuple coordinate (RA,DEC)')
    else:
        center_coord = SkyCoord(ra = center_coord[0], dec= center_coord[1], unit='deg', frame='icrs')

    # use only center quadrant
    imageshape = np.shape(image)
    cut = 1 - center_radius
    suby, subx = imageshape[0] * cut, imageshape[1] * cut
    center_source_inds = np.where((guesses['x'] >= subx) & (guesses['x'] <= imageshape[1] - subx) & (guesses['y'] >= suby) & (guesses['y'] <= imageshape[0] - suby))
    sub_guesses = guesses[center_source_inds]
    sub_skycoords = skycoords[center_source_inds]

    # Platescale calculation
    # create random combinations of stars
    all_combinations = list(compute_combinations(sub_guesses))
    if len(all_combinations) > 200:
        rand_inds = np.random.randint(low=0, high=len(all_combinations), size=200)
        combo_list = np.array(all_combinations)[rand_inds]
    else:
        combo_list = np.array(all_combinations)

    # gather the skycoord separations for all combinations
    seps = np.empty(len(combo_list))
    for i,c in enumerate(combo_list):
        star1 = sub_skycoords[c[0]]
        star2 = sub_skycoords[c[1]]

        sep = star1.separation(star2).mas
        seps[i] = sep

    # find the separations in pixel space on the image between all combinations
    pixseps = np.empty(len(combo_list))
    for i,c in enumerate(combo_list):
        star1 = sub_guesses[c[0]]
        star2 = sub_guesses[c[1]]

        xguess = star2['x'] - star1['x']
        yguess = star2['y'] - star1['y']
        
        (xoff, yoff), _ = measure_offset(image, xstar_guess=star1['x'], ystar_guess=star1['y'], xoffset_guess= xguess, yoffset_guess= yguess)

        pixsep = np.sqrt(np.power(xoff,2) + np.power(yoff,2))
        pixseps[i] = pixsep

    # estimate the platescale from each combination and find the mean
    platescales = seps / pixseps
    platescale = np.mean(platescales)

    # North angle calculation
    # find the true centerings of the sources in the image from the guesses and save into a table
    xs = np.empty(len(sub_guesses))
    ys = np.empty(len(sub_guesses))
    for i, (gx, gy) in enumerate(zip(sub_guesses['x'], sub_guesses['y'])):
        pf, fw, x, y = fakes.gaussfit2d(frame= image, xguess= gx, yguess=gy)
        xs[i] = x
        ys[i] = y

    sources = astropy.table.Table()
    sources['x'] = xs
    sources['y'] = ys

    # find the sky position angles between the center star and all others
    pa_sky = np.empty(len(sub_skycoords))
    for i, star in enumerate(sub_skycoords):
        pa = center_coord.position_angle(star).deg
        pa_sky[i] = pa

    # find the pixel position angles
    pa_pixel = np.empty(len(sub_guesses))
    for i, (x, y) in enumerate(zip(xs, ys)):
        pa = angle_between(((np.shape(image)[0]-1)//2, (np.shape(image)[1]-1)//2), (x,y))
        pa_pixel[i] = pa

    # find the difference between the measured and true positon angles
    offset = np.empty(len(sub_guesses))
    # locate a potential comparison with self
    if len(np.where((sub_skycoords.ra.value == center_coord.ra.value) & (sub_skycoords.dec.value == center_coord.dec.value))[0]) > 0:
        same_ind = np.where((sub_skycoords.ra.value == center_coord.ra.value) & (sub_skycoords.dec.value == center_coord.dec.value))[0][0]
    else:
        same_ind = None

    for i, (sky, pix) in enumerate(zip(pa_sky, pa_pixel)):
        if i != same_ind:
            numerator = np.sin(np.radians(sky - pix))
            denominator = np.cos(np.radians(sky - pix))
            north_offset = np.degrees(np.arctan(numerator / denominator))
            # if sky > pix:
            #     north_offset = sky - pix
            # else:
            #     north_offset = sky - pix + 360 
            offset[i] = north_offset

    # get rid of the comparison with self if it exists
    if same_ind != None:
        offset = np.delete(offset, same_ind)

    # use the median to avoid bias
    north_angle = np.mean(offset)
    
    return platescale, north_angle

def compute_boresight(image, source_info, target_coordinate, cal_properties):
    """ 
    Used to find the offset between the target and the center of the image.

    Args:
        image (numpy.ndarray): 2D array of image data
        source_info (astropy.table.Table): Estimated pixel positions of sources and true sky positions, must have column names 'x', 'y', 'RA', 'DEC'
        target_coordinate (tuple): 
            (float): RA coordinate of the target pointing
            (float): DEC coordinate of the target pointing
        cal_properties (tuple):
            (float): Platescale
            (float): North angle

    Returns:
        image_center_RA (float): RA coordinate of the center pixel
        image_center_DEC (float): Dec coordinate of the center pixel
    
    """
    if type(image) != np.ndarray:
        raise TypeError('Image must be 2D numpy.ndarray')
    
    if type(source_info) != astropy.table.Table:
        raise TypeError('source_info must be an astropy table with columns \'x\',\'y\',\'RA\',\'DEC\'')
    else:
        guesses = source_info
        skycoords = SkyCoord(ra = guesses['RA'], dec= guesses['DEC'], unit='deg', frame='icrs')

    if type(target_coordinate) != tuple:
        raise TypeError('target_coordinate must be tuple (RA,DEC)')

    if type(cal_properties) != tuple:
        raise TypeError('cal_properties must be tuple (platescale, north_angle)')

    # use only center quadrant
    imageshape = np.shape(image)
    quady, quadx = imageshape[0] // 4, imageshape[1] // 4
    center_source_inds = np.where((guesses['x'] >= quadx) & (guesses['x'] <= imageshape[1] - quadx) & (guesses['y'] >= quady) & (guesses['y'] <= imageshape[0] - quady))
    quad_guesses = guesses[center_source_inds]

    # create the predicted image header from found platescale and north angle
    vert_ang = np.radians(cal_properties[1])
    pc = np.array([[-np.cos(vert_ang), np.sin(vert_ang)], [np.sin(vert_ang), np.cos(vert_ang)]])
    cdmatrix = pc * (cal_properties[0] * 0.001) / 3600.

    new_hdr = {}
    new_hdr['CD1_1'] = cdmatrix[0,0]
    new_hdr['CD1_2'] = cdmatrix[0,1]
    new_hdr['CD2_1'] = cdmatrix[1,0]
    new_hdr['CD2_2'] = cdmatrix[1,1]
    new_hdr['CRPIX1'] = np.shape(image)[1] // 2
    new_hdr['CRPIX2'] = np.shape(image)[0] // 2
    new_hdr['CTYPE1'] = 'RA---TAN'
    new_hdr['CTYPE2'] = 'DEC--TAN'
    new_hdr['CDELT1'] = (cal_properties[0] * 0.001) / 3600
    new_hdr['CDELT2'] = (cal_properties[0] * 0.001) / 3600
    new_hdr['CRVAL1'] = target_coordinate[0]
    new_hdr['CRVAL2'] = target_coordinate[1]
    w = astropy.wcs.WCS(new_hdr)

    x_sky_to_pix, y_sky_to_pix = astropy.wcs.utils.skycoord_to_pixel(skycoords, wcs=w)
    x_predict, y_predict = x_sky_to_pix[center_source_inds], y_sky_to_pix[center_source_inds]

    # find offset between measured centers and predicted positions    
    image_centerings = np.zeros((len(quad_guesses), 2))
    boresights = np.zeros((len(quad_guesses), 2))
    searchrad = 5
    for i, (xg, yg) in enumerate(zip(quad_guesses['x'], quad_guesses['y'])):
        p, f, xi_center, yi_center = fakes.gaussfit2d(frame= image, xguess= xg, yguess= yg, searchrad=searchrad)
        x_off = xi_center - x_predict[i]
        y_off = yi_center - y_predict[i]
        boresights[i,:] = [x_off, y_off]
        image_centerings[i,:] = [xi_center, yi_center]

    # average all offsets in x,y directions [pix]
    boresight_x, boresight_y = np.mean(boresights[:,0]), np.mean(boresights[:,1])

    # convert back to corrected RA, DEC of target
    # image_center_RA = target_coordinate[0] - ((boresight_x * cal_properties[0]) * astropy.units.mas).to(astropy.units.deg).value
    # image_center_DEC = target_coordinate[1] - ((boresight_y * cal_properties[0]) * astropy.units.mas).to(astropy.units.deg).value

    # report the offsets instead of the new RA/DEC
    boresight_ra = ((boresight_x * cal_properties[0]) * astropy.units.mas).to(astropy.units.deg).value
    boresight_dec = ((boresight_y * cal_properties[0]) * astropy.units.mas).to(astropy.units.deg).value

    return boresight_ra, boresight_dec

def format_distortion_inputs(input_dataset, source_matches, ref_star_pos, position_error=None):
    ''' Function that formats the input data for the distortion map computation * must be run before compute_distortion *
    
    Args:
        input_dataset (corgidrp.data.dataset): corgidrp dataset object with images to compute the distortion from
        source_matches (list of astropy.table.Table() objects): List of length N for N frames in the input dataset. Tables must columns 'x','y','RA','DEC' as pixel locations and corresponding sky positons
        ref_star_pos (list of astropy.table.Table() objects): List of length N for N frames. Tables must have column names 'x', 'y', 'RA', 'DEC' for the position of the reference position to compute pairs with
        position_error (NoneType or int): If int, this is the uniform error value assumed for the offset between pairs of stars in both x and y
                        Should be changed later to accept non-uniform errors
        
    Returns:
        first_stars (np.array): 2D array of the (x, y) pixel positions for the first star in every star pair
        offsets (np.array): 2D array of the (delta_x, delta_y) values for each star from the first star position
        true_offsets (np.array): 2D array of the (delta_ra, delta_dec) offsets between the matched stars in the reference field
        errs (np.array): 2D array of the (x_err, y_err) error in the measured pixel positions
    '''
    ## Create arrays to store values in
    dxs, dys = np.array([]), np.array([])
    xerrs, yerrs = np.array([]), np.array([])
    firstxs, firstys = np.array([]), np.array([])
    seps, pas = np.array([]), np.array([])

    ## Loop over every frame in the dataset
    for frame_ind in range(len(input_dataset)):
        input_image = input_dataset[frame_ind].data

        # create all combinations of the target star with all others
        combo_list = range(len(source_matches[frame_ind]))
        skycoords = SkyCoord(ra= source_matches[frame_ind]['RA'], dec= source_matches[frame_ind]['DEC'], unit='deg', frame='icrs')
        target_coord = SkyCoord(ra= ref_star_pos[frame_ind]['RA'], dec= ref_star_pos[frame_ind]['DEC'], unit='deg', frame='icrs')
    
        for pair_ind in combo_list:
            # get the pixel offset
            star1 = ref_star_pos[frame_ind]
            star2 = source_matches[frame_ind][pair_ind]
    
            x_guess = star2['x'] - star1['x']
            y_guess = star2['y'] - star1['y']
        
            (dx, dy), (xfit_err, yfit_err, _) = measure_offset(input_image, star2['x'], star2['y'], x_guess, y_guess, guessflux=10000)
    
            # get the true sky offset [mas]
            true1 = target_coord
            true2 = skycoords[pair_ind]
        
            # get true sky separation and position angle
            true_sep = true1.separation(true2).mas
            true_pa = true1.position_angle(true2).deg
            
            dxs = np.append(dxs, dx)
            dys = np.append(dys, dy)
            firstxs = np.append(firstxs, star1['x'])
            firstys = np.append(firstys, star1['y'])
            seps = np.append(seps, true_sep)
            pas = np.append(pas, true_pa)
    
            if type(position_error) == type(None):
                xerrs = np.append(xerrs, xfit_err)
                yerrs = np.append(yerrs, yfit_err)
            else:
                xerrs = np.append(xerrs, position_error)
                yerrs = np.append(yerrs, position_error)
   
    # join arrays and reshape to (1, 2, N)
    offsets = np.array([dxs, dys])
    first_stars = np.array([firstxs, firstys])
    true_offsets = np.array([seps, pas])
    errs = np.array([xerrs, yerrs])

    return first_stars, offsets, true_offsets, errs

def compute_distortion(input_dataset, pos1, meas_offset, sky_offset, meas_errs, platescale, northangle, fitorder=3, initial_guess=None):
    ''' 
    Function that computes the legendre polynomial coefficients that describe the image distortion map * must run format_disotrtio_inputs() first *

    Args:
        input_dataset (corgidrp.data.Dataset): corgidrp dataset object with images to compute the distortion from
        pos1 (np.array): 2D array of the (x, y) pixel positions for the first star in every star pair
        meas_offset (np.array): 2D array of the (delta_x, delta_y) values for each star from the first star position
        sky_offset (np.array): 2D array of the (delta_ra, delta_dec) offsets between the matched stars in the reference field
        meas_errs (np.array): 2D array of the (x_err, y_err) error in the measured pixel positions
        platescale (float): Platescale value to use in computing distortion
        northangle (float): Northangle value to use in computing distortion 
        fitorder (int): The order of legendre polynomial to fit to the image distortion (default: 3)
        initial_guess (np.array): Initial guess of fitting parameters (legendre coefficients) length based on fitorder (2 * (fitorder+1)**2), (default: None)

    Returns:
        distortion_coeffs (tuple): The legendre coefficients (np.array) and polynomial order used for the fit (int)

    '''

    ## SET FITTING PARAMS
    # assume all images in dataset have the same shape
    input_image = input_dataset[0].data
    x0 = np.shape(input_image)[1] // 2
    y0 = np.shape(input_image)[0] // 2
    
    # define fitting params            
    fitparams = (fitorder + 1)**2
    
    # initial guesses for the legendre coeffs if none are passed
    if initial_guess is None:
        initial_guess = [0 for _ in range(fitorder+1)] + [500,] + [0 for _ in range(fitparams - fitorder - 2)] + [0,500] + [0 for _ in range(fitparams-2)]
    
    ## OPTIMIZE 
    # first_stars_, offsets_, true_offsets_, errs_ = first_stars, offsets, true_offsets, errs
    (distortion_coeffs, _) = optimize.leastsq(fit_distortion_solution, initial_guess, 
                                              args=(fitorder, platescale, 
                                                northangle, pos1, meas_offset, 
                                                sky_offset, meas_errs))

    return (distortion_coeffs, fitorder)
  
  
def boresight_calibration(input_dataset, field_path='JWST_CALFIELD2020.csv', field_matches=None, find_threshold=10, fwhm=7, mask_rad=1, 
                          comparison_threshold=50, search_rad=0.012, platescale_guess=21.8, platescale_tol=0.1, center_radius=0.9, 
                          frames_to_combine=None, find_distortion=False, fitorder=3, position_error=None, initial_dist_guess=None):
    """
    Perform the boresight calibration of a dataset.
    
    Args:
        input_dataset (corgidrp.data.Dataset): Dataset containing a images for astrometric calibration
        field_path (str): Full path to file with search field data (ra, dec, vmag, etc.) (default: 'JWST_CALFIELD2020.csv')
        field_matches (list of str or astropy.table.Table): List of full paths to files or astropy tables with calibration field matches for each image in the dataset (x, y, ra, dec), if single str the same filepath used for all frames,nif None, automated source matching is used (default: None)
        find_threshold (int): Number of stars to find (default 10)
        fwhm (float): Full width at half maximum of the stellar psf (default: 7, ~fwhm for a normal distribution with sigma=3)
        mask_rad (int): Radius of mask for stars [in fwhm] (default: 1)
        comparison_threshold (int): How many stars in the field to consider for the initial match (default: 50)
        search_rad (float): The radius [deg] around the target coordinate for creating a subfield to match image sources to
        platescale_guess (float): An initial guess for the platescale value (default: 21.8 [mas/ pixel])
        platescale_tol (float): A tolerance for finding source matches within a fraction of the initial plate scale guess (default: 0.1)
        center_radius (float): Percent of the image to compute plate scale and north angle from, centered around the image center (default: 0.9 -- ie: 90% of the image is used)
        frames_to_combine (int): The number of frames to combine in a dataset (default: None)
        find_distortion (boolean): Used to determine if distortion map coeffs will be computed (default: False)
        fitorder (int): The order of legendre polynomials used to fit the distortion map (default: 3)
        position_error (NoneType or int): If int, this is the uniform error value assumed for the offset between pairs of stars in both x and y
        initial_dist_guess (np.array): An initial guess of legendre coefficients used for fitting distortion, if None will use coeffs associated with no distortion (default: None)

    Returns:
        corgidrp.data.AstrometricCalibration: Astrometric Calibration data object containing image center coords in (RA,DEC), platescale, and north angle
        
    """
    # load in the data considering multiple frames in the data
    dataset = input_dataset.copy()

    # load in the source matches if automated source finder is not being used
    matched_sources_multiframe = []
    if field_matches is not None:
        if len(field_matches) == 1: # single str or astropy.table case
            for i in range(len(dataset)):
                if type(field_matches[0]) == str:
                    matched_sources = ascii.read(field_matches[0])
                    matched_sources_multiframe.append(matched_sources)
                else:
                    matched_sources_multiframe.append(field_matches[0])
        # elif len(field_matches) == len(dataset): # this needs to be if the len(field_matches >1)
        elif len(field_matches) > 1:  # unique matches for each frame case
            if len(field_matches) != len(dataset):
                raise TypeError('field_matches must be a single str/ astropy.table OR the same length as input_dataset')
            else:
                for i in range(len(field_matches)):
                    if type(field_matches[0]) == str:
                        matched_sources = ascii.read(field_matches[i])
                        matched_sources_multiframe.append(matched_sources)
                    else:
                        matched_sources_multiframe = field_matches
        # else:
        #     raise TypeError('field_matches must be a single str or the same length as input_dataset')

    # load in field data to refer to
    if field_path == 'JWST_CALFIELD2020.csv':
        full_field_path = os.path.join(os.path.dirname(__file__), "data", field_path)
        field_path = full_field_path

    # combine data frames if requested
    if frames_to_combine is not None:
        num_frames = len(input_dataset)
        data_array = []
        for frame in range(num_frames):
            data_array.append(input_dataset[frame].data)

        image_objects = []
        count = 0
        while count < num_frames:
            count += frames_to_combine
            if count >= num_frames:
                sub_array = data_array[count - frames_to_combine:]
                file_name = input_dataset[-1].filename
            else:
                sub_array = data_array[count - frames_to_combine: count]
                file_name = input_dataset[count].filename

            comb = np.mean(sub_array, axis=0)
            im = corgidrp.data.Image(comb, pri_hdr=input_dataset[count - frames_to_combine].pri_hdr, ext_hdr=input_dataset[0].ext_hdr)
            im.filename = file_name
            image_objects.append(im)
        
        dataset = corgidrp.data.Dataset(image_objects)

    # create a place to store all the calibration measurements
    astroms = []
    target_coord_tables = []

    hold_matches = []   # place to hold the auto-found source matches for each frame
    corrected_positions_boresight = []      # place to hold the corrected target position based on boresight offsets for each frame

    for i in range(len(dataset)):
        in_dataset = corgidrp.data.Dataset([dataset[i]])
        image = dataset[i].data

        # call the target coordinates from the image header
        target_coordinate = (dataset[i].pri_hdr['RA'], dataset[i].pri_hdr['DEC'])
        target_coord_tab = astropy.table.Table()
        target_coord_tab['x'] = [(np.shape(image)[1]-1) // 2]    # assume the target is at the center of the image
        target_coord_tab['y'] = [(np.shape(image)[0]-1) // 2]
        target_coord_tab['RA'] = [target_coordinate[0]]
        target_coord_tab['DEC'] = [target_coordinate[1]]
        target_coord_tables.append(target_coord_tab)
   
        # compute the calibration properties
        found_sources = find_source_locations(image, threshold=find_threshold, fwhm=fwhm, mask_rad=mask_rad)
        matched_sources = match_sources(dataset[i], found_sources, field_path, comparison_threshold=comparison_threshold, rad=search_rad, platescale_guess=platescale_guess, platescale_tol=platescale_tol)
        # if len(hold_matches) < 1:
        hold_matches.append(matched_sources)

        cal_properties = compute_platescale_and_northangle(image, source_info=matched_sources, center_coord=target_coordinate, center_radius=center_radius)
        ra, dec = compute_boresight(image, source_info=matched_sources, target_coordinate=target_coordinate, cal_properties=cal_properties)
        # calculate the corrected target position based on ra, dec offsets
        corr_ra, corr_dec = target_coordinate[0] - ra, target_coordinate[1] - dec
        corrected_positions_boresight.append([corr_ra, corr_dec])

        # return a single AstrometricCalibration data file
        astrom_data = np.array([corr_ra, corr_dec, cal_properties[0], cal_properties[0], cal_properties[1], ra, dec, np.inf, np.inf])
        astrom_cal = corgidrp.data.AstrometricCalibration(astrom_data, pri_hdr=dataset[i].pri_hdr, ext_hdr=dataset[i].ext_hdr, input_dataset=in_dataset)
        # change the filename here since the astrom_cals will be averaged later and arent individually saved ('_AST_CAL' will be added to filename twice otherwise)
        astrom_cal.filename = astrom_cal.filename.split("_AST_CAL")[0] + '.fits'
        astroms.append(astrom_cal)

    # average the calibration properties over all frames
    avg_ra = np.mean([astro.avg_offset[0] for astro in astroms])  # this is the average ra offset [deg]
    avg_dec = np.mean([astro.avg_offset[1] for astro in astroms])
    avg_platescale_x = np.mean([astro.platescale[0] for astro in astroms])
    avg_platescale_y = np.mean([astro.platescale[1] for astro in astroms])
    avg_northangle = np.mean([astro.northangle for astro in astroms])

    # compute the distortion map coeffs
    if find_distortion:
        # use the found matches for distortion
        first_stars, offsets, true_offsets, errs = format_distortion_inputs(input_dataset, source_matches=hold_matches, ref_star_pos=target_coord_tables, position_error=position_error)
        mean_xy_platescale = np.mean([avg_platescale_x, avg_platescale_y])
        distortion_coeffs, order = compute_distortion(input_dataset, first_stars, offsets, true_offsets, errs, platescale=mean_xy_platescale, northangle=avg_northangle, fitorder=fitorder, initial_guess=initial_dist_guess)
    else:
        # set default coeffs to produce zero distortion
        fitparams = (fitorder + 1)**2
        zero_dist = [0 for _ in range(fitorder+1)] + [500,] + [0 for _ in range(fitparams - fitorder - 2)] + [0,500] + [0 for _ in range(fitparams-2)]
        distortion_coeffs = np.array(zero_dist)
        order = fitorder

    # assume that the undithered image with original pointing position is the first frame in dataset
    corr_pos_ra, corr_pos_dec = corrected_positions_boresight[0]
    astromcal_data = np.concatenate((np.array([corr_pos_ra, corr_pos_dec, avg_platescale_x, avg_platescale_y, avg_northangle, avg_ra, avg_dec]), np.array(distortion_coeffs), np.array([order])), axis=0)

    astroms_dataset = corgidrp.data.Dataset(astroms)
    avg_cal = corgidrp.data.AstrometricCalibration(astromcal_data, pri_hdr=input_dataset[0].pri_hdr, ext_hdr=input_dataset[0].ext_hdr, input_dataset=astroms_dataset)
    # add the corrected RA/DEC for each frame to the ext_hdr
    for i, corr in enumerate(corrected_positions_boresight):
        name = 'F'+str(i)+'POS'
        avg_cal.ext_hdr[name] = tuple(corr)
        
    # update the history
    history_msg = "Boresight calibration completed"
    astrom_cal_dataset = corgidrp.data.Dataset([avg_cal])
    astrom_cal_dataset.update_after_processing_step(history_msg)
    ## history message should be added to the input dataset(?)
    input_dataset.update_after_processing_step(history_msg)

    return avg_cal


def create_circular_mask(shape_yx, center=None, r=None):
    """Creates a circular mask

    Args:
        shape_yx (list-like of int): 
        center (list of float, optional): Center of mask. Defaults to the 
            center of the array.
        r (float, optional): radius of mask. Defaults to the minimum distance 
            from the center to the edge of the array.

    Returns:
        np.array: boolean array with True inside the circle, False outside.
    """
    shape_yx = np.array(shape_yx)
    if center is None: # use the middle of the image
        center = (shape_yx-1) / 2
    if r is None: # use the smallest distance between the center and image walls
        r = min(center[0], center[1], shape_yx[0]-center[0], shape_yx[1]-center[1])

    Y, X = np.ogrid[:shape_yx[0], :shape_yx[1]]
    dist_from_center = np.sqrt((X - center[0])**2 + (Y-center[1])**2)

    mask = dist_from_center <= r
    return mask



import os
import glob
from pathlib import Path
import numpy as np
from astropy.io import fits
import pytest
import re
import copy

from corgidrp import default_cal_dir
import corgidrp.fluxcal as fluxcal
import corgidrp.nd_filter_calibration as nd_filter_calibration
import corgidrp.l2b_to_l3 as l2b_tol3
import corgidrp.data as data 
from corgidrp.data import Dataset
from corgidrp.data import Image
import corgidrp.mocks as mocks
from corgidrp.data import Image, NDFilterSweetSpotDataset

# ---------------------------------------------------------------------------
# Global variables and constants
# ---------------------------------------------------------------------------
BRIGHT_STARS = ['Vega']
DIM_STARS = ['TYC 4424-1286-1',
             'GSC 02581-02323']

# takes a long time to run with all stars
#BRIGHT_STARS = ['109 Vir', 'Vega', 'Eta Uma', 'Lam Lep']
#DIM_STARS = ['TYC 4433-1800-1', 'TYC 4205-1677-1', 'TYC 4212-455-1', 'TYC 4209-1396-1',
#            'TYC 4413-304-1', 'UCAC3 313-62260', 'BPS BS 17447-0067', 'TYC 4424-1286-1',
#             'GSC 02581-02323', 'TYC 4207-219-1']

DIM_EXPTIME = 10.0
BRIGHT_EXPTIME = 5.0
FWHM = 3
FILTER_USED = '3C'
INPUT_OD = 2.25
CAL_FACTOR = 0.8
OD_RASTER_THRESHOLD = 0.1
OD_TEST_TOLERANCE = 0.2
FILESAVE = True
ADD_GAUSS = False
BACKGROUND = 3
PHOT_METHOD = "Aperture"
FLUX_OR_IRR = 'irr'

if PHOT_METHOD == "Aperture":
    PHOT_ARGS = {
        "encircled_radius": 7,
        "frac_enc_energy": 1,
        "method": "subpixel",
        "subpixels": 10,
        "background_sub": True,
        "r_in": 5,
        "r_out": 10,
        "centering_method": "xy",
        "centroid_roi_radius": 5
    }
elif PHOT_METHOD == "Gaussian":
    PHOT_ARGS = {
        "fwhm": 3,
        "fit_shape": None,
        "background_sub": True,
        "r_in": 5,
        "r_out": 10,
        "centering_method": 'xy',
        "centroid_roi_radius": 5
    }

# ---------------------------------------------------------------------------
# Functions to generate mocks
# ---------------------------------------------------------------------------
def mock_dim_dataset_files(dim_exptime, filter_used, cal_factor, save_mocks, output_path=None, 
                           background_val=0, add_gauss_noise_val=False):
    """
    Generate and save mock dim dataset files for specified exposure time and filter.

    Parameters:
        dim_exptime (float): Exposure time for the simulated images.
        filter_used (str): Filter used for the observations.
        cal_factor (float): Calibration factor applied to the images.
        save_mocks (bool): Whether to save the generated mock images.
        output_path (str, optional): Directory path to save the images. Defaults to the current working directory.
        background_val (int, optional): Background value to be added to the images. Defaults to 0.
        add_gauss_noise_val (bool, optional): Whether to add Gaussian noise to the images. Defaults to False.

    Returns:
        list: A list of generated flux images for the dim stars.
    """
    if save_mocks:
        output_path = output_path or os.getcwd()
    else:
        output_path = output_path or os.getcwd()
    os.makedirs(output_path, exist_ok=True)
    dim_star_images = []
    for star_name in DIM_STARS:
        dim_star_flux = nd_filter_calibration.compute_expected_band_irradiance(star_name, filter_used)
        flux_image = mocks.create_flux_image(
            dim_star_flux, FWHM, cal_factor, filter=filter_used, fpamname="HOLE", target_name=star_name,
            fsm_x=0, fsm_y=0, exptime=dim_exptime, filedir=output_path,
            platescale=21.8,
            background=background_val,
            add_gauss_noise=add_gauss_noise_val,
            noise_scale=1.0, file_save=True
        )
        dim_star_images.append(flux_image)
    return dim_star_images


def mock_bright_dataset_files(bright_exptime, filter_used, OD, cal_factor, save_mocks, output_path=None, 
                              background_val=0, add_gauss_noise_val=False):
    """
    Generate and save mock bright dataset files for specified exposure time and filter.

    Parameters:
        bright_exptime (float): Exposure time for the simulated images.
        filter_used (str): Filter used for the observations.
        OD (float): The OD used for the observations.
        cal_factor (float): Calibration factor applied to the images.
        save_mocks (bool): Whether to save the generated mock images.
        output_path (str, optional): Directory path to save the images. Defaults to the current working directory.
        background_val (int, optional): Background value to be added to the images. Defaults to 0.
        add_gauss_noise_val (bool, optional): Whether to add Gaussian noise to the images. Defaults to False.

    Returns:
        list: A list of generated flux images for the dim stars.
    """
    if save_mocks:
        output_path = output_path or os.getcwd()
    else:
        output_path = output_path or os.getcwd()
    os.makedirs(output_path, exist_ok=True)
    ND_transmission = 10 ** (-OD)
    bright_star_images = []
    for star_name in BRIGHT_STARS:
        for dy in [-10, 0, 10]:
            for dx in [-10, 0, 10]:
                bright_star_flux = nd_filter_calibration.compute_expected_band_irradiance(star_name, 
                                                                                          filter_used)
                attenuated_flux = bright_star_flux * ND_transmission
                flux_image = mocks.create_flux_image(
                    attenuated_flux, FWHM, cal_factor, filter=filter_used, fpamname="ND225", target_name=star_name,
                    fsm_x=dx, fsm_y=dy, exptime=bright_exptime, filedir=output_path,
                    platescale=21.8,
                    background=background_val,
                    add_gauss_noise=add_gauss_noise_val,
                    noise_scale=1.0, file_save=True
                )
                bright_star_images.append(flux_image)
    return bright_star_images


def mock_clean_entry(bright_dataset):
    # TO DO: eventually add other processing steps that would produce the 
    # appropriate level input file
    return bright_dataset[0]


def mock_transformation_matrix(output_dir):
    # TO DO: use Sergi's FPAM to EXCAM transformation matrix product once it is
    # ready
    transformation_matrix_file = os.path.join(output_dir, "fpam_to_excam.fits")
    dummy_matrix = np.eye(2)
    hdu = fits.PrimaryHDU(dummy_matrix)
    hdu.writeto(transformation_matrix_file, overwrite=True)
    return transformation_matrix_file

# ---------------------------------------------------------------------------
# Pytest fixtures
# ---------------------------------------------------------------------------
# These fixtures generate and cache the mocks (which is time consuming) for the
# entire module
@pytest.fixture(scope="module", params=[(0, False), (BACKGROUND, ADD_GAUSS)])
def bg_settings(request):
    return request.param

@pytest.fixture(scope="module")
def dim_files_cached(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp("dim_dataset")
    print(f"Generating cached dim dataset in {tmp_dir}")
    files = mock_dim_dataset_files(DIM_EXPTIME, FILTER_USED, CAL_FACTOR, save_mocks=False, output_path=str(tmp_dir))
    return files

@pytest.fixture(scope="module")
def bright_files_cached(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp("bright_dataset")
    print(f"Generating cached bright dataset in {tmp_dir}")
    files = mock_bright_dataset_files(BRIGHT_EXPTIME, FILTER_USED, INPUT_OD, CAL_FACTOR, save_mocks=False, output_path=str(tmp_dir))
    return files

@pytest.fixture(scope="module")
def stars_dataset_cached(bright_files_cached, dim_files_cached):
    combined_files = bright_files_cached + dim_files_cached
    # TO DO: May eventually need to add other processing steps to get the mocks
    # to a representative input state
    return l2b_tol3.divide_by_exptime(Dataset(combined_files))

@pytest.fixture(scope="module")
def dim_dir(tmp_path_factory):
    """
    Creates a temporary directory, populates it with the mock dim FITS files,
    then returns the directory path.
    """
    tmp_dir = tmp_path_factory.mktemp("dim_dir_fixture")
    mock_dim_dataset_files(
        DIM_EXPTIME,
        FILTER_USED,
        CAL_FACTOR,
        save_mocks=True,
        output_path=str(tmp_dir),
        background_val=0,
        add_gauss_noise_val=False
    )
    return str(tmp_dir)

@pytest.fixture
def output_dir(tmp_path):
    out = tmp_path / "output"
    out.mkdir()
    print(f"Created temporary output directory: {out}")
    return str(out)

# ---------------------------------------------------------------------------
# Test functions using pytest
# ---------------------------------------------------------------------------
def test_nd_filter_calibration_object(stars_dataset_cached):
    print("**Testing ND filter calibration object generation and expected headers**")
    # don't want the datasets to get overwritten for subsequent tests
    ds_copy = copy.deepcopy(stars_dataset_cached)
    results = nd_filter_calibration.create_nd_filter_cal(
        ds_copy, OD_RASTER_THRESHOLD, PHOT_METHOD, FLUX_OR_IRR, PHOT_ARGS, 
        fluxcal_factor = None)
    
    results.save(filedir=default_cal_dir)

    nd_files = [fn for fn in os.listdir(default_cal_dir) if fn.endswith('_NDF_CAL.fits')]
    assert nd_files, "No NDFilterOD files were generated."
    with fits.open(os.path.join(default_cal_dir, nd_files[0])) as hdul:
        primary_hdr = hdul[0].header
        ext_hdr = hdul[1].header
        assert primary_hdr.get('SIMPLE') is True, "Primary header missing or SIMPLE not True."
        assert ext_hdr.get('FPAMNAME') is not None, "Missing FPAMNAME keyword."
        assert ext_hdr.get('FPAM_H') is not None, "Missing FPAM_H keyword."
        assert ext_hdr.get('FPAM_V') is not None, "Missing FPAM_V keyword."
        assert ext_hdr.get('CFAMNAME') is not None, "Missing CFAMNAME keyword."


def test_output_filename_convention(stars_dataset_cached):
    print("**Testing output filename naming conventions**")
    
    # Make a copy of the dataset and retrieve expected values.
    ds_copy = copy.deepcopy(stars_dataset_cached)

    # Construct the expected filename from the last input dataset filename.
    expected_filename = re.sub('_L[0-9].', '_NDF_CAL', stars_dataset_cached[-1].filename)
    full_expected_path = os.path.join(default_cal_dir, expected_filename)

    # Create the calibration product
    results = nd_filter_calibration.create_nd_filter_cal(
        ds_copy, OD_RASTER_THRESHOLD, PHOT_METHOD, FLUX_OR_IRR, PHOT_ARGS,
        fluxcal_factor=None
    )
    results.save(filedir=default_cal_dir)
    
    assert os.path.exists(full_expected_path), (
        f"Expected file {expected_filename} not found in {default_cal_dir}."
    )
    print("The nd_filter_calibration product file exists and meets the expected naming convention.")


def test_average_od_within_tolerance(stars_dataset_cached):
    print("**Testing computed OD within tolerance**")
    ds_copy = copy.deepcopy(stars_dataset_cached)
    results = nd_filter_calibration.create_nd_filter_cal(
        ds_copy, OD_RASTER_THRESHOLD, PHOT_METHOD, FLUX_OR_IRR, PHOT_ARGS, 
        fluxcal_factor = None)
    ods = results.data
    avg_od = np.mean(ods[:, 0])
    std_od = np.std(ods[:,0])
    results_hdr = results.ext_hdr
    od_flag = results_hdr.get('ODFLAG')

    assert abs(avg_od - INPUT_OD) < OD_TEST_TOLERANCE, (
        f"Avg OD {avg_od} deviates more than {OD_TEST_TOLERANCE}"
    )
    if std_od < OD_RASTER_THRESHOLD:
        assert not od_flag, f"Low OD variation but flag is set"

    else:
        assert od_flag, f"High OD variation but flag is not set"


@pytest.mark.parametrize("phot_method", ["Aperture", "Gaussian"])
def test_nd_filter_calibration_phot_methods(stars_dataset_cached, phot_method):
    if phot_method == "Aperture":
        phot_args = {
            "encircled_radius": 7,
            "frac_enc_energy": 1,
            "method": "subpixel",
            "subpixels": 10,
            "background_sub": True,
            "r_in": 5,
            "r_out": 10,
            "centering_method": "xy",
            "centroid_roi_radius": 5
        }
    else:
        phot_args = {
            "fwhm": 3,
            "fit_shape": None,
            "background_sub": True,
            "r_in": 5,
            "r_out": 10,
            "centering_method": "xy",
            "centroid_roi_radius": 5
        }
    print(f"**Testing ND calibration with photometry method: {phot_method}**")
    ds_copy = copy.deepcopy(stars_dataset_cached)
    results = nd_filter_calibration.create_nd_filter_cal(
        ds_copy, OD_RASTER_THRESHOLD, phot_method, FLUX_OR_IRR, phot_args, 
        fluxcal_factor = None)
    ods = results.data
    avg_od = np.mean(ods[:, 0])
    assert abs(avg_od - INPUT_OD) < OD_TEST_TOLERANCE, (
        f"Method {phot_method}: OD mismatch"
    )


@pytest.mark.parametrize("test_od", [1.0, 3.0])
def test_multiple_nd_levels(dim_dir, output_dir, test_od):
    print(f"**Testing multiple ND levels with input OD = {test_od}**")
    bright_mocks_dir = os.path.join(output_dir, f"mock_OD{test_od}")
    bright_images = mock_bright_dataset_files(
        BRIGHT_EXPTIME, FILTER_USED, test_od, CAL_FACTOR, save_mocks=True, output_path=bright_mocks_dir
    )

    dim_filepaths = glob.glob(os.path.join(dim_dir, "*")) # use cached dim images
    dim_images = [Image(path) for path in dim_filepaths]
    combined_files = bright_images + dim_images
    combined_dataset = l2b_tol3.divide_by_exptime(Dataset(combined_files))

    results = nd_filter_calibration.create_nd_filter_cal(
        combined_dataset, OD_RASTER_THRESHOLD, PHOT_METHOD, FLUX_OR_IRR, PHOT_ARGS, 
        fluxcal_factor = None)
    ods = results.data
    avg_od = np.mean(ods[:, 0])
    assert abs(avg_od - test_od) < 0.2, (
        f"test_od={test_od}, got {avg_od} for target."
    )

@pytest.mark.parametrize("phot_method", ["Aperture", "Gaussian"])
def test_nd_filter_calibration_with_fluxcal(dim_dir, stars_dataset_cached, phot_method):
    """
    1) Takes a star image from stars_dataset_cached
    2) Derives a flux calibration factor using calibrate_fluxcal_aper or calibrate_fluxcal_gauss2d
    3) Passes that flux calibration factor to create_nd_filter_cal
    4) Checks the resulting ND filter calibration to ensure it ran as expected
    """
    print(f"**Testing with FluxcalFactor input using method {phot_method}**")
    # Pick images without the ND filter in
    dim_filepaths = glob.glob(os.path.join(dim_dir, "*.fits"))
    assert len(dim_filepaths) > 0, f"No FITS files found in {dim_dir}"
    
    dim_images = [Image(path) for path in dim_filepaths]

    # Convert list of Image objects into a Dataset
    dim_dataset = l2b_tol3.divide_by_exptime(Dataset(dim_images))

    # 1) Generate a flux calibration object from the single image
    if phot_method == "Aperture":
        phot_kwargs = {
            "encircled_radius": 7,
            "frac_enc_energy": 1.0,
            "method": "subpixel",
            "subpixels": 10,
            "background_sub": True,
            "r_in": 5,
            "r_out": 10,
            "centering_method": "xy",
            "centroid_roi_radius": 5
        }
        fluxcal_obj = fluxcal.calibrate_fluxcal_aper(
            dim_dataset, 
            flux_or_irr="irr", 
            phot_kwargs=phot_kwargs
        )
    else:
        phot_kwargs = {
            "fwhm": 3,
            "fit_shape": None,
            "background_sub": True,
            "r_in": 5,
            "r_out": 10,
            "centering_method": "xy",
            "centroid_roi_radius": 5
        }
        fluxcal_obj = fluxcal.calibrate_fluxcal_gauss2d(
            dim_dataset, 
            flux_or_irr="irr", 
            phot_kwargs=phot_kwargs
        )

    results = nd_filter_calibration.create_nd_filter_cal(
        stars_dataset_cached,
        OD_RASTER_THRESHOLD,
        PHOT_METHOD,
        FLUX_OR_IRR,
        PHOT_ARGS,
        fluxcal_factor=fluxcal_obj
    )

    # 3) Check that the calibration worked
    ods = results.data
    assert ods.shape[0] > 0, "No data returned from ND calibration."
    avg_od = np.mean(ods[:, 0])

    # Check nominal OD
    assert abs(avg_od - INPUT_OD) < 0.3, (
        f"Expected OD near {INPUT_OD}, got {avg_od}"
    )

    print(f"ND filter calibration with fluxcal completed. OD={avg_od} (mean).")


@pytest.mark.parametrize("aper_radius", [5, 10])
def test_aperture_radius_sensitivity(stars_dataset_cached, aper_radius):
    print(f"**Testing aperture radius sensitivity: radius = {aper_radius}**")
    phot_args = {
        "encircled_radius": aper_radius,
        "frac_enc_energy": 1,
        "method": "subpixel",
        "subpixels": 5,
        "background_sub": True,
        "r_in": 5,
        "r_out": 10,
        "centering_method": "xy",
        "centroid_roi_radius": 5
    }
    ds_copy = copy.deepcopy(stars_dataset_cached)
    results = nd_filter_calibration.create_nd_filter_cal(
        ds_copy, OD_RASTER_THRESHOLD, "Aperture", "irr", phot_args, 
        fluxcal_factor = None)
    ods = results.data
    avg_od = np.mean(ods[:, 0])
    assert abs(avg_od - INPUT_OD) < 0.3, (
        f"AperRadius={aper_radius}: OD mismatch for target."
    )


def test_background_effect(tmp_path):
    """
    Generate two sets of mocks (one without background, one with background)
    and compare the calibration results. We expect that the overall OD values
    remain similar while the error increases when background is added.
    """
    # Create dim star mocks for two modes.
    dim_dir_no = tmp_path / "dim_no"
    dim_dir_bg = tmp_path / "dim_bg"
    dim_dir_no.mkdir(exist_ok=True)
    dim_dir_bg.mkdir(exist_ok=True)
    dim_files_no = mock_dim_dataset_files(DIM_EXPTIME, FILTER_USED, CAL_FACTOR, save_mocks=False,
                                      output_path=str(dim_dir_no), background_val=0, add_gauss_noise_val=False)
    dim_files_bg = mock_dim_dataset_files(DIM_EXPTIME, FILTER_USED, CAL_FACTOR, save_mocks=False,
                                      output_path=str(dim_dir_bg), background_val=BACKGROUND, add_gauss_noise_val=ADD_GAUSS)

    # Create bright star mocks for two modes.
    bright_dir_no = tmp_path / "bright_no"
    bright_dir_bg = tmp_path / "bright_bg"
    bright_dir_no.mkdir(exist_ok=True)
    bright_dir_bg.mkdir(exist_ok=True)
    bright_files_no = mock_bright_dataset_files(BRIGHT_EXPTIME, FILTER_USED, INPUT_OD, CAL_FACTOR, save_mocks=False,
                                                output_path=str(bright_dir_no), background_val=0, add_gauss_noise_val=False)
    bright_files_bg = mock_bright_dataset_files(BRIGHT_EXPTIME, FILTER_USED, INPUT_OD, CAL_FACTOR, save_mocks=False,
                                                output_path=str(bright_dir_bg), background_val=BACKGROUND, add_gauss_noise_val=ADD_GAUSS)
    
    combined_no = dim_files_no + bright_files_no
    combined_bg = dim_files_bg + bright_files_bg
    ds_no = l2b_tol3.divide_by_exptime(Dataset(combined_no))
    ds_bg = l2b_tol3.divide_by_exptime(Dataset(combined_bg))
    
    output_directory = str(tmp_path / "output")
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)
    
    results_no = nd_filter_calibration.create_nd_filter_cal(
        ds_no, OD_RASTER_THRESHOLD, PHOT_METHOD, FLUX_OR_IRR, PHOT_ARGS, 
        fluxcal_factor = None)
    results_bg = nd_filter_calibration.create_nd_filter_cal(
        ds_bg, OD_RASTER_THRESHOLD, PHOT_METHOD, FLUX_OR_IRR, PHOT_ARGS,
        fluxcal_factor = None
    )

    ods_no = results_no.data
    avg_od_no = np.mean(ods_no[:, 0])
    stdev_od_no = np.std(ods_no[:,0])
    ods_bg = results_bg.data
    avg_od_bg = np.mean(ods_bg[:, 0])
    stdev_od_bg = np.std(ods_bg[:,0])
    
    print(f"Avg OD no-bg = {avg_od_no}, stdev no-bg = {stdev_od_no}; Avg OD bg = {avg_od_bg}, stdev bg = {stdev_od_bg}")
    # The overall OD should be similar, within a small tolerance.
    assert abs(avg_od_no - avg_od_bg) < 0.1, f"OD should not differ drastically between background subtraction and no background subtraction modes."


def test_calculate_od_at_new_location(output_dir):
    """
    Test calculate_od_at_new_location with:
      1) A real NDFilterSweetSpotDataset containing Nx3 data = [OD, x, y]
      2) A mock clean_frame_entry with a star centroid at a known location
      3) Known FPAM offsets in headers
      4) An identity transformation matrix file
    """

    # Create a small Nx3 sweet spot array, each row is [OD, x, y].
    # Interpolate at the center (5,5).
    # The OD values at corners are 2.0, 3.0, 4.0, 5.0 => a bilinear interpolation at (5,5) => 3.5.
    sweetspot_data = np.array([
        [2.0,  0.0,  0.0],   # OD=2.0 at (x=0,y=0)
        [3.0,  0.0, 10.0],   # OD=3.0 at (x=0,y=10)
        [4.0, 10.0,  0.0],   # OD=4.0 at (x=10,y=0)
        [5.0, 10.0, 10.0]    # OD=5.0 at (x=10,y=10)
    ], dtype=float)

    # Create a fake input dataset to set the filename
    input_prihdr, input_exthdr = mocks.create_default_L2b_headers()
    fake_input_image = Image(sweetspot_data, pri_hdr=input_prihdr, ext_hdr=input_exthdr)
    fake_input_image.filename = f"CGI_{input_prihdr['VISITID']}_{data.format_ftimeutc(input_exthdr['FTIMEUTC'])}_L2b.fits"
    fake_input_dataset = Dataset(frames_or_filepaths=[fake_input_image, fake_input_image])

    # Build the NDFilterSweetSpotDataset
    ndcal_prihdr, ndcal_exthdr = mocks.create_default_calibration_product_headers()
    ndcal_exthdr["FPAM_H"] = 0.0
    ndcal_exthdr["FPAM_V"] = 0.0
    nd_sweetspot_dataset = NDFilterSweetSpotDataset(data_or_filepath=sweetspot_data, pri_hdr=ndcal_prihdr, ext_hdr=ndcal_exthdr,
                                                    input_dataset=fake_input_dataset)
 
    # Create an identity transformation matrix FITS file in output_dir
    transformation_matrix_file = mock_transformation_matrix(output_dir)

    # Make a 5x5 mock 'clean_frame_entry' with a star at (2,2) => centroid (2,2)
    # Shift it by (3,3) => final location (5,5).
    clean_image_data = np.zeros((5, 5), dtype=float)
    clean_image_data[2, 2] = 100.0  # star pixel
    cframe_prihdr, cframe_exthdr = mocks.create_default_L2b_headers()
    cframe_exthdr["FPAM_H"] = 3.0
    cframe_exthdr["FPAM_V"] = 3.0
    clean_frame_entry = Image(data_or_filepath=clean_image_data, pri_hdr=cframe_prihdr, 
                              ext_hdr=cframe_exthdr)

    # Call the function under test
    interpolated_od = nd_filter_calibration.calculate_od_at_new_location(
        clean_frame_entry=clean_frame_entry,
        transformation_matrix_file=transformation_matrix_file,
        ndsweetspot_dataset=nd_sweetspot_dataset
    )

    # Expect the final location = (2+3, 2+3) = (5,5).
    # Bilinear interpolation of corners: (2,3,4,5) at center => 3.5
    expected_value = 3.5
    assert abs(interpolated_od - expected_value) < 1e-6, (
        f"Expected OD={expected_value}, got {interpolated_od}"
    )
    print(
        f"test_calculate_od_at_new_location_nd_sweetspot_real_class PASSED: "
        f"interpolated OD={interpolated_od}"
    )


BRIGHT_CACHE_DIR = "/Users/jmilton/Github/corgidrp/corgidrp/data/nd_filter_mocks/bright"
DIM_CACHE_DIR = "/Users/jmilton/Github/corgidrp/corgidrp/data/nd_filter_mocks/dim"

def main():
    """
    this is for julia testing
    """

    # Make sure the directories exist
    os.makedirs(BRIGHT_CACHE_DIR, exist_ok=True)
    os.makedirs(DIM_CACHE_DIR, exist_ok=True)

    # 1) See if any .fits exist for bright stars
    bright_files = glob.glob(os.path.join(BRIGHT_CACHE_DIR, "*.fits"))
    if bright_files:
        print(f"[main()] Using cached bright .fits from:\n  {BRIGHT_CACHE_DIR}")
    else:
        print(f"[main()] No cached bright .fits found. Generating new mocks in:\n  {BRIGHT_CACHE_DIR}")
        bright_files = mock_bright_dataset_files(
            BRIGHT_EXPTIME,
            FILTER_USED,
            INPUT_OD,
            CAL_FACTOR,
            save_mocks=False,         # no real saving beyond memory?
            output_path=BRIGHT_CACHE_DIR,
        )

    # 2) See if any .fits exist for dim stars
    dim_files = glob.glob(os.path.join(DIM_CACHE_DIR, "*.fits"))
    if dim_files:
        print(f"[main()] Using cached dim .fits from:\n  {DIM_CACHE_DIR}")
    else:
        print(f"[main()] No cached dim .fits found. Generating new mocks in:\n  {DIM_CACHE_DIR}")
        dim_files = mock_dim_dataset_files(
            DIM_EXPTIME,
            FILTER_USED,
            CAL_FACTOR,
            save_mocks=False,
            output_path=DIM_CACHE_DIR,
        )

    # 3) Combine them into stars_dataset_cached
    combined_files = bright_files + dim_files
    stars_dataset_cached = l2b_tol3.divide_by_exptime(Dataset(combined_files))

    # 4) Create an "output" subdirectory for storing calibration products, etc.
    output_dir = os.path.join(os.path.dirname(DIM_CACHE_DIR), "output")
    os.makedirs(output_dir, exist_ok=True)

    # Need a "tmp" directory for test_background_effect (which expects tmp_path).
    background_tmp_dir = Path(os.path.join(os.path.dirname(DIM_CACHE_DIR), "tmp"))
    background_tmp_dir.mkdir(exist_ok=True)

    def run_test(test_func, *args, **kwargs):
        try:
            test_func(*args, **kwargs)
            print(f"{test_func.__name__} PASSED")
        except Exception as exc:
            print(f"{test_func.__name__} FAILED: {exc}")
            raise  

    print("\n========== BEGIN TESTS ==========")

    run_test(test_nd_filter_calibration_object, stars_dataset_cached)
    run_test(test_output_filename_convention, stars_dataset_cached)
    run_test(test_average_od_within_tolerance, stars_dataset_cached)

    for method in ["Aperture", "Gaussian"]:
        run_test(test_nd_filter_calibration_phot_methods, stars_dataset_cached, method)

    for test_od in [1.0, 3.0]:
        run_test(test_multiple_nd_levels, DIM_CACHE_DIR, output_dir, test_od)

    for aper_radius in [5, 10]:
        run_test(test_aperture_radius_sensitivity, stars_dataset_cached, aper_radius)

    run_test(test_background_effect, background_tmp_dir)

    run_test(test_nd_filter_calibration_with_fluxcal, DIM_CACHE_DIR, stars_dataset_cached, "Gaussian")

    test_calculate_od_at_new_location(output_dir)

    print("All tests PASSED")


if __name__ == "__main__":
    main()



import os
import pytest
import numpy as np
import astropy.io.fits as fits
import corgidrp
import corgidrp.mocks as mocks
import corgidrp.detector as detector
from corgidrp.mocks import create_default_L2a_headers
from corgidrp.data import Image, Dataset, DetectorParams
import corgidrp.caldb as caldb
from corgidrp.darks import build_trad_dark

np.random.seed(123)

data = np.ones([1024,1024]) * 2
err = np.zeros([1024,1024])
err1 = np.ones([1024,1024])
err2 = err1.copy()
err3 = np.ones([1,1024,1024]) * 0.5
dq = np.zeros([1024,1024], dtype = int)
dq1 = dq.copy()
dq1[0,0] = 1
prhd, exthd = create_default_L2a_headers()
errhd = fits.Header()
errhd["CASE"] = "test"
dqhd = fits.Header()
dqhd["CASE"] = "test"

data_3d = np.ones([2,1024,1024]) * 2
err_3d = np.zeros([2,1024,1024])
err1_3d = np.ones([2,1024,1024])
err2_3d = err1_3d.copy()
err3_3d = np.ones([2,1,1024,1024]) * 0.5
dq_3d = np.zeros([2,1024,1024], dtype = int)
dq1_3d = dq_3d.copy()
dq1_3d[0,0,0] = 1

old_err_tracking = corgidrp.track_individual_errors
# use default parameters
detector_params = DetectorParams({})

def test_err_dq_creation():
    """
    test the initialization of error and dq attributes of the Image class including saving and loading

    Test assuming track individual error terms is on
    """
    corgidrp.track_individual_errors = True

    image1 = Image(data,pri_hdr = prhd, ext_hdr = exthd)
    assert hasattr(image1, "err")
    assert hasattr(image1, "dq")
    assert image1.data.shape == data.shape
    assert image1.data.shape == image1.err.shape[-2:]
    assert image1.err.shape == (1, 1024, 1024)
    assert image1.data.shape == image1.dq.shape
    #test the initial error and dq headers
    assert hasattr(image1, "err_hdr")
    assert hasattr(image1, "dq_hdr")
    assert np.mean(image1.dq) == 0
    image1.save(filename='test_image1.fits')

    image2 = Image(data,pri_hdr = prhd, ext_hdr = exthd, err = err, dq = dq1, err_hdr = errhd, dq_hdr = dqhd)
    # test the user defined error and dq headers
    assert image2.err_hdr["CASE"] == errhd["CASE"]
    assert image2.dq_hdr["CASE"] == dqhd["CASE"]
    #check the correct saving and loading of fits files
    image2.save(filename='test_image2.fits')
    image3 = Image('test_image2.fits')
    assert np.array_equal(image3.data, image2.data)
    assert np.array_equal(image3.err, image2.err)
    assert np.array_equal(image3.dq, image2.dq)
    assert image3.err_hdr["CASE"] == errhd["CASE"]
    assert image3.dq_hdr["CASE"] == dqhd["CASE"]

    #check the overwriting of err and dq parameters of the fits extensions arrays
    image_test = Image('test_image2.fits', err = err, dq = dq)
    assert np.array_equal(image_test.dq, dq)
    assert np.array_equal(image_test.err[0,:,:], err)
    #test 3d input error array
    image_test2 = Image('test_image2.fits', err = err3)
    assert np.array_equal(image_test2.err, err3)

def test_err_dq_copy():
    """
    test the copying of the err and dq attributes

    Runs assuming tracking individual errors
    """
    corgidrp.track_individual_errors = True

    image2 = Image('test_image2.fits')
    image3 = image2.copy()
    assert np.array_equal(image3.data, image2.data)
    assert np.array_equal(image3.err, image2.err)
    assert np.array_equal(image3.dq, image2.dq)
    #check the copying of the headers of the extensions
    assert image2.err_hdr == image3.err_hdr
    assert image2.dq_hdr == image3.dq_hdr

def test_add_error_term():
    """
    test the add_error_term function

    Runs assuming tracking individual errors
    """
    corgidrp.track_individual_errors = True

    image1 = Image('test_image1.fits')
    image1.add_error_term(err1, "error_noid")
    assert image1.err[0,0,0] == err1[0,0]
    image1.add_error_term(err2, "error_nuts")
    assert image1.err.shape == (3,1024,1024)
    assert image1.err[0,0,0] == np.sqrt(err1[0,0]**2 + err2[0,0]**2)
    image1.save(filename="test_image0.fits")

    image_test = Image('test_image0.fits')
    assert np.array_equal(image_test.dq, dq)
    assert np.array_equal(image_test.err, image1.err)
    assert image_test.err.shape == (3,1024,1024)
    assert image_test.err_hdr["Layer_1"] == "combined_error"
    assert image_test.err_hdr["Layer_2"] == "error_noid"
    assert image_test.err_hdr["Layer_3"] == "error_nuts"

    image_3d = Image(data_3d,prhd,exthd,err_3d,dq_3d,errhd,dqhd)
    image_3d.add_error_term(err1_3d, "error_noid")
    assert image_3d.err[0,0,0,0] == err_3d[0,0,0]
    image_3d.add_error_term(err2_3d, "error_nuts")
    assert image_3d.err.shape == (3,2,1024,1024)
    assert image_3d.err[0,0,0,0] == np.sqrt(err1_3d[0,0,0]**2 + err2_3d[0,0,0]**2)
    image_3d.save(filename="test_image3d.fits")

    image_test_3d = Image('test_image3d.fits')
    assert np.array_equal(image_test_3d.dq, dq_3d)
    assert np.array_equal(image_test_3d.err, image_3d.err)
    assert image_test_3d.err.shape == (3,2,1024,1024)
    assert image_test_3d.err_hdr["Layer_1"] == "combined_error"
    assert image_test_3d.err_hdr["Layer_2"] == "error_noid"
    assert image_test_3d.err_hdr["Layer_3"] == "error_nuts"

def test_err_dq_dataset():
    """
    test the behavior of the err and data arrays in the dataset

    Runs assuming tracking individual errors
    """
    corgidrp.track_individual_errors = True

    dataset = Dataset(["test_image1.fits", "test_image2.fits"])
    assert np.array_equal(dataset[0].data, dataset[1].data)
    assert np.array_equal(dataset[0].err, dataset[1].err)
    assert not np.array_equal(dataset[0].dq, dataset[1].dq)
    assert dataset.all_data.shape[0] == len(dataset)
    assert dataset.all_err.ndim == 4
    assert dataset.all_dq.ndim == 3
    assert dataset.all_err.shape[2] == err.shape[0]
    assert dataset.all_dq.shape[2] == dq.shape[1]

def test_get_masked_data():
    """
    test the masked array
    """
    image2 = Image('test_image2.fits')
    masked_data = image2.get_masked_data()
    assert masked_data.data[0,1] == 2
    #check that pixel 0,0 is masked and not considered
    assert masked_data.mask[0,0] == True
    assert masked_data.mean()==2
    assert masked_data.sum()==image2.data.sum()-2

def test_err_adderr_notrack():
    """
    test the initialization of error and adding errors when we are not tracking
    individual errors. There should always only be a single 2-D map.
    """
    corgidrp.track_individual_errors = False

    image1 = Image(data,pri_hdr = prhd, ext_hdr = exthd)
    assert hasattr(image1, "err")
    assert image1.data.shape == data.shape
    assert image1.data.shape == image1.err.shape[-2:]
    assert image1.err.shape == (1, 1024, 1024)
    #test the initial error and dq headers
    assert hasattr(image1, "err_hdr")

    image1.add_error_term(err1, "error_noid")
    assert image1.err[0,0,0] == err1[0,0]
    image1.add_error_term(err2, "error_nuts")
    assert image1.err.shape == (1,1024,1024)
    assert image1.err[0,0,0] == np.sqrt(err1[0,0]**2 + err2[0,0]**2)

def test_read_many_errors_notrack():
    """
    Check that we can successfully discard errors when reading in a frame with multiple errors

    """
    corgidrp.track_individual_errors = False

    image_test = Image('test_image0.fits')
    assert image_test.err.shape == (1,1024,1024)
    assert image_test.err_hdr["Layer_1"] == "combined_error"
    with pytest.raises(KeyError):
        assert image_test.err_hdr["Layer_2"] == "error_noid"
    with pytest.raises(KeyError):
        assert image_test.err_hdr["Layer_3"] == "error_nuts"

def test_err_array_sizes():
    '''
    Check that we're robust to 2D error arrays

    Creates a dark calibration and then forces the error array to be 2D

    Makes sure that we're robust to that.
    '''

    ##### Create a master Dark #####
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    dark_dataset = mocks.create_dark_calib_files(filedir=datadir)
    dark_frame = build_trad_dark(dark_dataset, detector_params, detector_regions=None, full_frame=True)

    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    # clean out directory of old cals .fits files
    for filename in os.listdir(calibdir):
        if filename.endswith(".fits"):
            os.remove(os.path.join(calibdir, filename))
            
    dark_filename = "sim_dark_calib.fits"
    if not os.path.exists(calibdir):
            os.mkdir(calibdir)
    dark_frame.save(filedir=calibdir, filename=dark_filename)

    ##### Scan the caldb ##### - This tests for previous bug that darks weren't in the right format.
    testcaldb_filepath = os.path.join(calibdir, "test_caldb.csv")
    testcaldb = caldb.CalDB(filepath=testcaldb_filepath)
    testcaldb.scan_dir_for_new_entries(calibdir)

    ##### Force it to be 2D ##### - This tests to maks sure we're robust to 2D error arrays in general
    dark_frame.err = np.ones(dark_frame.data.shape)
    dark_frame.save(filedir=calibdir, filename=dark_filename)
    testcaldb.scan_dir_for_new_entries(calibdir)

def teardown_module():
    """
    Runs automatically at the end. ONLY IN PYTEST

    Removes new FITS files and restores track individual error setting.
    """
    for i in range(3):
        os.remove('test_image{0}.fits'.format(i))

    corgidrp.track_individual_errors = old_err_tracking

# for debugging. does not run with pytest!!
if __name__ == '__main__':
    test_err_array_sizes()
    test_err_dq_creation()
    test_err_dq_copy()
    test_add_error_term()
    test_err_dq_dataset()
    test_get_masked_data()
    test_err_adderr_notrack()
    test_read_many_errors_notrack()
    test_err_array_sizes()

    for i in range(3):
        os.remove('test_image{0}.fits'.format(i))
    os.remove('test_image3d.fits')


from corgidrp.mocks import create_psfsub_dataset,create_ct_cal, create_default_L4_headers
from corgidrp.klip_fm import measure_noise
from corgidrp.l3_to_l4 import do_psf_subtraction
from corgidrp.l4_to_tda import compute_flux_ratio_noise
import corgidrp.data as data
import corgidrp.mocks as mocks

import pytest
import numpy as np

## Helper functions/quantities

iwa_lod = 3.
owa_lod = 9.7
d = 2.36 #m
lam = 573.8e-9 #m
pixscale_arcsec = 0.0218
fwhm_mas = 1.22 * lam / d * 206265 * 1000
fwhm_pix = fwhm_mas * 0.001 / pixscale_arcsec
sig_pix = fwhm_pix / (2 * np.sqrt(2. * np.log(2.)))
iwa_pix = iwa_lod * lam / d * 206265 / pixscale_arcsec
owa_pix = owa_lod * lam / d * 206265 / pixscale_arcsec

# Mock CT calibration

annuli = 1
subsections = 1
movement = 1
calibrate_flux = False

st_amp = 100.
noise_amp = 1e-3
pl_contrast = 0.0
rolls = [0,15.,0,0]
numbasis = [1]

max_thrupt_tolerance = 1 + (noise_amp * (2*fwhm_pix)**2)


def test_expected_flux_ratio_noise():

    # RDI
    mode = 'RDI'
    nsci, nref = (1,1)
 
    st_amp = 100.
    noise_amp = 1e-3
    pl_contrast = 0. # No planet
    rolls = [0,15.,0,0]
    numbasis = [1,2]
    mock_sci_rdi,mock_ref_rdi = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast)

    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    
    psfsub_dataset_rdi = do_psf_subtraction(mock_sci_rdi,ctcal,
                                reference_star_dataset=mock_ref_rdi,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=None,
                                do_crop=False,
                                measure_klip_thrupt=True,
                                measure_1d_core_thrupt=True)
    
    # make unocculted star 
    x = np.arange(psfsub_dataset_rdi[0].data.shape[-1])
    y = np.arange(psfsub_dataset_rdi[0].data.shape[-2])
    X,Y = np.meshgrid(x,y)
    XY = np.vstack([X.ravel(),Y.ravel()])
    def gauss_spot(xy, A, x0, y0, sx, sy):
        (x, y) = xy
        return A*np.e**(-((x-x0)**2/(2*sx**2) + (y-y0)**2/(2*sy**2)))
    star_amp = 100
    x0 = 15
    y0 = 17
    sig_x = 4
    sig_y = 4
    FWHM_star = 2*np.sqrt(2*np.log(2))*sig_x
    # expected flux of star, same for each frame of input dataset to compute_flux_ratio_noise:
    # integral under Gaussian times ND transmission
    Fs_expected = np.pi*star_amp*FWHM_star**2/(4*np.log(2)) * 1e-2
    star_PSF = np.reshape(gauss_spot(XY,star_amp,x0,y0,sig_x,sig_y), X.shape)
    # add some noise to the star 
    np.random.seed(987)
    star_PSF += np.random.poisson(lam=star_PSF.mean(), size=star_PSF.shape)
    prihdr, exthdr = create_default_L4_headers()
    star_image = data.Image(star_PSF, prihdr, exthdr)
    star_dataset = data.Dataset([star_image for i in range(len(psfsub_dataset_rdi))])
    # fake an ND calibration:
    nd_x, nd_y = np.meshgrid(np.linspace(0, 100, 5), np.linspace(0, 100, 5))
    nd_x = nd_x.ravel()
    nd_y = nd_y.ravel()
    nd_od = np.ones(nd_y.shape) * 1e-2
    pri_hdr, ext_hdr = mocks.create_default_L2b_headers()
    nd_cal = data.NDFilterSweetSpotDataset(np.array([nd_od, nd_x, nd_y]).T, pri_hdr=pri_hdr,
                                      ext_hdr=ext_hdr)
    
    # now see what the step function gives, with and without a supplied star location guess:
    frn_dataset_nostarloc = compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, halfwidth=3)
    frn_dataset_starloc = compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, unocculted_star_loc=np.array([[17],[15]]), halfwidth=3)

    for frn_dataset in [frn_dataset_nostarloc, frn_dataset_starloc]:
        for frame in frn_dataset:
            flux_ratio_noise = frame.hdu_list['FRN_CRV'].data
            frn_seps = flux_ratio_noise[0]
            klip = frame.hdu_list['KL_THRU'].data
            separations = klip[0,:,0]
            klip_tp = klip[1:,:,0]
            klip_fwhms = frame.hdu_list['KL_THRU'].data[1:,:,1]
            core_tp = frame.hdu_list['CT_THRU'].data[1]
            annular_noise = measure_noise(frame, separations, hw=3)
            noise_amp = annular_noise.T
            # expected planet flux: Guassian integral
            Fp_expected = np.pi*noise_amp*klip_fwhms**2/(4*np.log(2))
            # expected flux ratio noise, adjusting for KLIP and core throughputs
            frn_expected = (Fp_expected/Fs_expected)/(core_tp*klip_tp)
            flux_ratio_noise = frame.hdu_list['FRN_CRV'].data
            frn_seps = flux_ratio_noise[0]
            assert np.array_equal(frn_seps, separations)
            assert np.allclose(flux_ratio_noise[2:], frn_expected, rtol=0.05)
            assert 'FRN_CRV' in frn_dataset[0].ext_hdr["HISTORY"][-1]
            assert frn_dataset[0].hdu_list['FRN_CRV'].header['BUNIT'] == "Fp/Fs"

    # last 2 separations below close to what they are in the KLIP extension, and increased the length from 2 separations in KLIP extension to 3 here to make sure that can be interpolated and handled
    requested_separations = np.array([26.5, 6.4, 14.8]) 
    frn_dataset_rseps = compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, requested_separations=requested_separations, halfwidth=3)
    for frame in frn_dataset_rseps:
        flux_ratio_noise = frame.hdu_list['FRN_CRV'].data
        frn_seps = flux_ratio_noise[0]
        klip = frame.hdu_list['KL_THRU'].data
        klip_tp = klip[1:,:,0]
        klip_fwhms = frame.hdu_list['KL_THRU'].data[1:,:,1]
        core_tp = frame.hdu_list['CT_THRU'].data[1]
        annular_noise = measure_noise(frame, requested_separations, hw=3)
        # the last 2 separation values should be close to what is expected from KLIP non-interpolated values
        noise_amp = annular_noise.T[:,1:]
        # expected planet flux: Guassian integral
        Fp_expected = np.pi*noise_amp*klip_fwhms**2/(4*np.log(2))
        # expected flux ratio noise, adjusting for KLIP and core throughputs
        frn_expected = (Fp_expected/Fs_expected)/(core_tp*klip_tp)
        flux_ratio_noise = frame.hdu_list['FRN_CRV'].data
        frn_seps = flux_ratio_noise[0]
        assert np.array_equal(frn_seps, requested_separations)
        # the last 2 separation values should be close to what is expected from KLIP non-interpolated values
        assert np.allclose(flux_ratio_noise[2:,1:], frn_expected, rtol=0.05)

    # halfwidth from the KLIP extension data's 2 separations used below
    frn_dataset_hw = compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, halfwidth=None)
    frn_dataset_true_hw = compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, halfwidth=8.41973985/2)
    frn_hw = frn_dataset_hw.frames[0].hdu_list['FRN_CRV'].data
    frn_true_hw = frn_dataset_true_hw.frames[0].hdu_list['FRN_CRV'].data
    assert np.array_equal(frn_hw, frn_true_hw)
    
    #test inputs
    # number of frames in input_dataset and unocculted_star_dataset should be equal
    bad_star_dataset = data.Dataset([star_image for i in range(len(psfsub_dataset_rdi)+2)])
    with pytest.raises(ValueError):
        compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, bad_star_dataset)
    # requested_separations includes at least 1 value outside the range covered by the KLIP extension header, so extrapolation is used
    with pytest.warns(UserWarning):
        compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, requested_separations=np.array([4, 13, 17]))
    # bad halfwidth input value
    with pytest.raises(ValueError):
        compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, halfwidth=0)
    # Halfwidth is wider than half the minimum spacing between separation values
    with pytest.warns(UserWarning):
        compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, halfwidth=100)
    # star location guess outside array bounds
    with pytest.raises(ValueError):
        compute_flux_ratio_noise(psfsub_dataset_rdi, nd_cal, star_dataset, unocculted_star_loc=np.array([[50],[101]]))
    

if __name__ == '__main__':  
    test_expected_flux_ratio_noise()

    pass


import os
import pickle
import numpy as np
import astropy.time as time
import corgidrp
import corgidrp.data as data


expected_rowreadtime = 0.0002235
expected_fwc_em = 100000.
expected_fwc_pp = 90000.
expected_kgain = 8.7

def test_default_params():
    """
    Tests creation of the default file
    """
    default_detparams = data.DetectorParams({}, date_valid=time.Time("2023-11-01 00:00:00", scale='utc'))

    assert default_detparams.params['ROWREADT'] == expected_rowreadtime
    assert default_detparams.params['FWC_EM_E'] == expected_fwc_em
    assert default_detparams.params['FWC_PP_E'] == expected_fwc_pp
    assert default_detparams.params['KGAINPAR'] == expected_kgain

def test_hashing():
    """
    Tests the hash function produces the expected results
    """
    default_detparams = data.DetectorParams({}, date_valid=time.Time("2023-11-01 00:00:00", scale='utc'))
    default_detparams_2 = data.DetectorParams({}, date_valid=time.Time("2023-11-01 00:00:00", scale='utc'))

    assert default_detparams.get_hash() == default_detparams_2.get_hash()

    new_detparams = data.DetectorParams({'FWC_EM_E' : 200000}, date_valid=time.Time("2023-11-01 00:00:00", scale='utc'))
    assert default_detparams.get_hash() != new_detparams.get_hash()

def test_pickling():
    """
    Test detector params can be pickled
    """
    filename = os.path.join(corgidrp.default_cal_dir, "DetectorParams_2023-11-01T00:00:00.000.fits")
    default_detparams = data.DetectorParams(filename)
        
    # check they can be pickled (for CTC operations)
    pickled = pickle.dumps(default_detparams)
    pickled_detparams = pickle.loads(pickled)
    assert np.all((default_detparams.data == pickled_detparams.data))


import os
import pytest
import numpy as np
import astropy.io.fits as fits
import corgidrp.data as data
import corgidrp.mocks as mocks
import corgidrp.combine as combine

img1 = np.ones([100, 100])
err1 = np.ones([100, 100])
dq = np.zeros([100, 100], dtype = int)
prhd, exthd = mocks.create_default_L1_headers()
errhd = fits.Header()
errhd["CASE"] = "test"
dqhd = fits.Header()
dqhd["CASE"] = "test"


def test_mean_combine_subexposures():
    """
    Test mean combine of subexposures
    """

    image1 = data.Image(img1, err=err1, dq=dq, pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    dataset = data.Dataset([image1, image2, image3, image4])

    combined_dataset = combine.combine_subexposures(dataset, 2)

    assert(len(combined_dataset) == 2)
    assert(np.all(combined_dataset[0].data == 2))
    assert(np.all(combined_dataset[0].err == pytest.approx(np.sqrt(2))))
    assert(np.all(combined_dataset[0].dq == 0))

    assert combined_dataset[0].ext_hdr['DRPNFILE'] == 2
    assert combined_dataset[0].ext_hdr['FILE0'] in ["2.fits", "1.fits"]
    assert combined_dataset[0].ext_hdr['FILE1'] in ["2.fits", "1.fits"]
    assert combined_dataset[0].ext_hdr['NUM_FR'] == 2

    assert combined_dataset[1].ext_hdr['DRPNFILE'] == 2
    assert combined_dataset[1].ext_hdr['FILE0'] in ["3.fits", "4.fits"]
    assert combined_dataset[1].ext_hdr['FILE1'] in ["3.fits", "4.fits"]
    assert combined_dataset[1].ext_hdr['NUM_FR'] == 2
    # combine again
    combined_dataset_2 = combine.combine_subexposures(combined_dataset, 2)
     
    assert(len(combined_dataset_2) == 1)
    assert(np.all(combined_dataset_2[0].data == 4))
    assert(np.all(combined_dataset_2[0].err == pytest.approx(2)))
    assert(np.all(combined_dataset_2[0].dq == 0))

    assert combined_dataset_2[0].ext_hdr['DRPNFILE'] == 4
    assert combined_dataset_2[0].ext_hdr['FILE0'] in ["2.fits", "1.fits", "3.fits", "4.fits"]
    assert combined_dataset_2[0].ext_hdr['FILE1'] in ["2.fits", "1.fits", "3.fits", "4.fits"]
    assert combined_dataset_2[0].ext_hdr['FILE2'] in ["2.fits", "1.fits", "3.fits", "4.fits"]
    assert combined_dataset_2[0].ext_hdr['FILE3'] in ["2.fits", "1.fits", "3.fits", "4.fits"]

def test_mean_combine_subexposures_without_scaling():
    """
    Test mean combine of subexposures for case where num_frames_scaling=False
    """

    image1 = data.Image(img1, err=err1, dq=dq, pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    dataset = data.Dataset([image1, image2, image3, image4])

    combined_dataset = combine.combine_subexposures(dataset, 2, num_frames_scaling=False)

    # Check that data and error values are not scaled up
    assert(len(combined_dataset) == 2)
    assert(np.all(combined_dataset[0].data == 1))
    assert(np.all(combined_dataset[0].err == pytest.approx(1/np.sqrt(2))))
    assert(np.all(combined_dataset[0].dq == 0))

    # combine again
    combined_dataset_2 = combine.combine_subexposures(combined_dataset, 2, num_frames_scaling=False)
     
    assert(len(combined_dataset_2) == 1)
    assert(np.all(combined_dataset_2[0].data == 1))
    assert(np.all(combined_dataset_2[0].err == pytest.approx((1/np.sqrt(2))/np.sqrt(2))))
    assert(np.all(combined_dataset_2[0].dq == 0))

def test_mean_combine_subexposures_with_bad():
    """
    Test mean combine of subexposures with bad pixels
    """
    # use copies since we are going to modify their values
    image1 = data.Image(np.copy(img1), err=np.copy(err1), dq=np.copy(dq), 
                        pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    # (0,0) has one bad frame
    image1.dq[0][0] = 1
    # (0,1) has both pixels bad
    image1.dq[0][1] = 1
    image2.dq[0][1] = 1

    dataset = data.Dataset([image1, image2, image3, image4])

    combined_dataset = combine.combine_subexposures(dataset, 2)

    assert(len(combined_dataset) == 2)
    # the pixel with one bad pixel should have same value but higher error
    assert combined_dataset[0].data[0][0] == 2
    assert combined_dataset[0].err[0][0][0] == pytest.approx(2)
    assert combined_dataset[0].dq[0][0] == 0 # 0 because one of the two frames had a good value
    # compare against a pixel without any bad pixels
    assert combined_dataset[1].data[0][0] == 2
    assert combined_dataset[1].err[0][0][0] == pytest.approx(np.sqrt(2))
    assert combined_dataset[1].dq[0][0] == 0

    # the pixel with two bad pixels should be nan
    assert np.isnan(combined_dataset[0].data[0][1])
    assert np.isnan(combined_dataset[0].err[0][0][1])
    assert combined_dataset[0].dq[0][1] == 1

def test_median_combine_subexposures():
    """
    Test median combine of subexposures. And tests default case where num_frames_per_group isn't specified. 
    """

    image1 = data.Image(img1, err=err1, dq=dq, pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    dataset = data.Dataset([image1, image2, image3, image4])

    combined_dataset = combine.combine_subexposures(dataset, collapse="median")

    assert(len(combined_dataset) == 1)
    assert(np.all(combined_dataset[0].data == 4))
    assert(np.all(combined_dataset[0].err == pytest.approx(np.sqrt(2*np.pi))))
    assert(np.all(combined_dataset[0].dq == 0))

def test_median_combine_subexposures_with_bad():
    """
    Test median combine of subexposures with bad pixels over multiple combinations
    """
    # use copies since we are going to modify their values
    image1 = data.Image(np.copy(img1), err=np.copy(err1), dq=np.copy(dq), 
                        pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    # (0,0) has one bad frame
    image1.dq[0][0] = 1
    # (0,1) has both pixels bad
    image1.dq[0][1] = 1
    image2.dq[0][1] = 1

    dataset = data.Dataset([image1, image2, image3, image4])

    combined_dataset = combine.combine_subexposures(dataset, 2, collapse="median")

    assert(len(combined_dataset) == 2)
    # the pixel with one bad pixel should have same value but higher error. In both cases the error should be inflated by np.sqrt(np.pi/2) compared to mean error.
    assert combined_dataset[0].data[0][0] == 2
    assert combined_dataset[0].err[0][0][0] == pytest.approx(2 * np.sqrt(np.pi/2))
    assert combined_dataset[0].dq[0][0] == 0 # 0 because one of the two frames had a good value
    # compare against a pixel without any bad pixels
    assert combined_dataset[1].data[0][0] == 2
    assert combined_dataset[1].err[0][0][0] == pytest.approx(np.sqrt(2) * np.sqrt(np.pi/2))
    assert combined_dataset[1].dq[0][0] == 0

    # the pixel with two bad pixels should be nan
    assert np.isnan(combined_dataset[0].data[0][1])
    assert np.isnan(combined_dataset[0].err[0][0][1])
    assert combined_dataset[0].dq[0][1] == 1

    # combine again
    combined_dataset_2 = combine.combine_subexposures(combined_dataset, 2, collapse="median")
     
    assert(len(combined_dataset_2) == 1)
    assert(np.all(combined_dataset_2[0].data == 4))

    # error for pixel with no bad pixels in original data (i.e. most pixels in data)
    assert combined_dataset_2[0].err[0][5][0] == pytest.approx(np.pi)

    # error for pixel with one bad pixel in original data (i.e. no nans after first combination)
    assert combined_dataset_2[0].err[0][0][0] == pytest.approx(0.5 * np.pi * np.sqrt(6))

    # error for pixel with two bad pixels in original data (i.e. 1 nan after first combination)
    assert combined_dataset_2[0].err[0][0][1] == pytest.approx(np.pi * np.sqrt(2))

    assert(np.all(combined_dataset_2[0].dq == 0))

def test_combine_different_values():
    """
    Test whether the function correctly combines different values.
    """

    # use copies since we are going to modify their values
    image1 = data.Image(np.copy(img1), err=np.copy(err1), dq=np.copy(dq), 
                        pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    # (0,0) has different values in each frame. Some are bad pixels.
    image1.data[0][0] = 5
    image2.data[0][0] = 6
    image3.data[0][0] = 9
    image4.data[0][0] = 19

    image2.dq[0][0] = 1
    image4.dq[0][0] = 1

    # (0,1) is a bad pixel in every frame
    image1.dq[0][1] = 1
    image2.dq[0][1] = 1
    image3.dq[0][1] = 1
    image4.dq[0][1] = 1

    dataset = data.Dataset([image1, image2, image3, image4])

    combined_dataset = combine.combine_subexposures(dataset, collapse="median")

    assert(len(combined_dataset) == 1)

    # Most pixels had good values of 1 in all frames
    assert combined_dataset[0].data[0][2] == 4
    assert combined_dataset[0].err[0][0][2] == pytest.approx(2*np.sqrt(np.pi/2))

    # (0,0) has a different median value calculated ignoring nans
    assert combined_dataset[0].data[0][0] == 7 * 4 # median value scaled by number of images
    assert combined_dataset[0].err[0][0][0] == pytest.approx(2 * np.sqrt(np.pi))

    # (0,1) is a nan
    assert np.isnan(combined_dataset[0].data[0][1])
    assert np.isnan(combined_dataset[0].err[0][0][1])

    # the updated bad pixel map only contains one bad pixel (i.e. the pixel for which there were no good values)
    assert combined_dataset[0].dq[0][0] == 0
    assert combined_dataset[0].dq[0][1] == 1

def test_not_divisible():
    """
    Tests that function correctly fails when the length of the dataset is not divisible by num_frames_per_group.
    """

    image1 = data.Image(img1, err=err1, dq=dq, pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    dataset = data.Dataset([image1, image2, image3, image4])

    with pytest.raises(ValueError):
        combined_dataset = combine.combine_subexposures(dataset, 3) # Should fail as 4 % 3 != 0

def test_invalid_collapse():
    """
    Tests that function correctly fails when collapse type is not valid.
    """

    image1 = data.Image(img1, err=err1, dq=dq, pri_hdr = prhd, ext_hdr = exthd)
    image1.filename = "1.fits"
    image2 = image1.copy()
    image2.filename = "2.fits"
    image3 = image1.copy()
    image3.filename = "3.fits"
    image4 = image1.copy()
    image4.filename = "4.fits"

    dataset = data.Dataset([image1, image2, image3, image4])

    with pytest.raises(ValueError):
        combined_dataset = combine.combine_subexposures(dataset, collapse="invalid_option")

if __name__ == "__main__":
    test_mean_combine_subexposures()
    test_mean_combine_subexposures_with_bad()
    test_median_combine_subexposures()


import os
import numpy as np
import pytest
import corgidrp
import corgidrp.caldb as caldb
import corgidrp.data as data
import corgidrp.mocks as mocks

datadir = os.path.join(os.path.dirname(__file__), "simdata")
calibdir = os.path.join(os.path.dirname(__file__), "testcalib")

if not os.path.exists(datadir):
    os.mkdir(datadir)
if not os.path.exists(calibdir):
    os.mkdir(calibdir)

testcaldb_filepath = os.path.join(calibdir, "test_caldb.csv")

# make some fake test data to use
np.random.seed(456)
dark_dataset = mocks.create_dark_calib_files()
master_dark = data.Dark(dark_dataset[0].data, dark_dataset[0].pri_hdr, dark_dataset[0].ext_hdr, dark_dataset)
# save master dark to disk to be loaded later
master_dark.save(filedir=calibdir, filename="mockdark.fits")

def test_caldb_init():
    """
    Tests that caldb has been initialized. It has to be if it's being imported.
    """
    assert caldb.initialized


def test_caldb_create_default():
    """
    Test caldb creation when no filepath is passed in (uses default path)
    """
    # remove any stranded testcaldb if needed
    if os.path.exists(testcaldb_filepath):
        os.remove(testcaldb_filepath)
    assert(not os.path.exists(testcaldb_filepath))

    # modify default path so we don't mess up the real thing
    old_path = corgidrp.caldb_filepath
    corgidrp.caldb_filepath = testcaldb_filepath

    # create the caldb and check it's saved to disk and empty
    testcaldb = caldb.CalDB()
    assert(testcaldb.filepath == testcaldb_filepath)
    assert(os.path.exists(testcaldb_filepath))
    assert(len(testcaldb._db.index) == 0)

    # remove db and restore path
    os.remove(testcaldb_filepath)
    corgidrp.caldb_filepath = old_path


def test_caldb_custom_filepath():
    """
    Test caldb creation when filepath is passed in (should be an edge case)
    """
    # remove any stranded testcaldb if needed
    if os.path.exists(testcaldb_filepath):
        os.remove(testcaldb_filepath)
    assert(not os.path.exists(testcaldb_filepath))

    # create the caldb and check it's saved to disk and empty
    testcaldb = caldb.CalDB(filepath=testcaldb_filepath)
    assert(testcaldb.filepath == testcaldb_filepath)
    assert(testcaldb.filepath != corgidrp.caldb_filepath)
    assert(os.path.exists(testcaldb_filepath))
    assert(len(testcaldb._db.index) == 0)

    # remove db and restore path
    os.remove(testcaldb_filepath)

def test_caldb_insert_and_remove():
    """
    Tests the ability to add and remove an entry successfully
    """
    # remove any stranded testcaldb if needed
    if os.path.exists(testcaldb_filepath):
        os.remove(testcaldb_filepath)
    assert(not os.path.exists(testcaldb_filepath))

    # create custom caldb for testing
    testcaldb = caldb.CalDB(filepath=testcaldb_filepath)
    assert(len(testcaldb._db.index) == 0)

    # add dark file into database and check that the values are stored correctly
    testcaldb.create_entry(master_dark)
    assert(len(testcaldb._db.index) == 1)
    assert(testcaldb._db['NAXIS1'][0] == master_dark.data.shape[1])
    orig_exptime = master_dark.ext_hdr['EXPTIME']
    assert(testcaldb._db['EXPTIME'][0] == orig_exptime)

    # test update
    master_dark.ext_hdr['EXPTIME'] = 2*orig_exptime
    testcaldb.create_entry(master_dark)
    assert(len(testcaldb._db.index) == 1)
    assert(testcaldb._db['NAXIS1'][0] == master_dark.data.shape[1])
    assert(testcaldb._db['EXPTIME'][0] == 2*orig_exptime)

    # test remove
    testcaldb.remove_entry(master_dark)
    assert(len(testcaldb._db.index) == 0)

    # reset everything
    master_dark.ext_hdr['EXPTIME'] = orig_exptime
    os.remove(testcaldb_filepath)

def test_get_calib():
    """
    Tests ability to load a calibration file from disk
    """
    # remove any stranded testcaldb if needed
    if os.path.exists(testcaldb_filepath):
        os.remove(testcaldb_filepath)
    assert(not os.path.exists(testcaldb_filepath))

    # create custom caldb for testing
    testcaldb = caldb.CalDB(filepath=testcaldb_filepath)
    assert(len(testcaldb._db.index) == 0)

    # add dark file into database and check that the values are stored correctly
    testcaldb.create_entry(master_dark)
    assert(len(testcaldb._db.index) == 1)

    # grab the only dark in the caldb, so it should be the one we put in
    auto_dark = testcaldb.get_calib(dark_dataset[2], data.Dark)
    assert(auto_dark.filepath == master_dark.filepath)

    with pytest.raises(ValueError):
        _ = testcaldb.get_calib(dark_dataset[2], data.DetectorNoiseMaps)

    # make a second one dark
    master_dark_2 = data.Dark(dark_dataset[1].data, dark_dataset[1].pri_hdr, dark_dataset[0].ext_hdr, dark_dataset)
    # save master dark to disk to be loaded later
    master_dark_2.save(filedir=calibdir, filename="mockdark2.fits")
    testcaldb.create_entry(master_dark_2)

    # test that with no input data, we get the most recent dark
    auto_dark_2 = testcaldb.get_calib(None, data.Dark)
    assert(auto_dark_2.filepath == master_dark_2.filepath)
        
    # reset everything
    os.remove(testcaldb_filepath)



def test_caldb_scan():
    """
    Tests ability to scan a folder to look for calibration files
    """
    # remove any stranded testcaldb if needed
    if os.path.exists(testcaldb_filepath):
        os.remove(testcaldb_filepath)
    assert(not os.path.exists(testcaldb_filepath))

    # create custom caldb for testing
    testcaldb = caldb.CalDB(filepath=testcaldb_filepath)
    assert(len(testcaldb._db.index) == 0)

    # there should be no calibration files in "./simdata", just unprocessed data
    testcaldb.scan_dir_for_new_entries(datadir)
    assert(len(testcaldb._db.index) == 0)

    # there should be at least the master dark in "./testcalib"
    testcaldb.scan_dir_for_new_entries(calibdir)
    assert(len(testcaldb._db.index) > 0)

    # reset everything
    os.remove(testcaldb_filepath)

if __name__ == "__main__":
    test_caldb_init()
    test_get_calib()
    test_caldb_create_default()
    test_caldb_custom_filepath()
    test_caldb_insert_and_remove()
    test_caldb_scan()